# -*- coding: utf-8 -*-
"""Attributes processor for span attributes."""
import datetime
import enum
import inspect
import json
from dataclasses import is_dataclass
from typing import Any

from pydantic import BaseModel

from ..message import Msg


def _to_serializable(
    obj: Any,
) -> Any:
    """Convert an object to a JSON serializable type.

    Args:
        obj (`Any`):
            The object to be converted to JSON serializable.

    Returns:
        `Any`:
            The converted JSON serializable object
    """

    # Handle primitive types first
    if isinstance(obj, (str, int, bool, float, type(None))):
        res = obj

    elif isinstance(obj, (list, tuple, set, frozenset)):
        res = [_to_serializable(x) for x in obj]

    elif isinstance(obj, dict):
        res = {str(key): _to_serializable(val) for (key, val) in obj.items()}

    elif isinstance(obj, (Msg, BaseModel)) or is_dataclass(obj):
        res = repr(obj)

    elif inspect.isclass(obj) and issubclass(obj, BaseModel):
        res = repr(obj)

    elif isinstance(obj, (datetime.date, datetime.datetime, datetime.time)):
        res = obj.isoformat()

    elif isinstance(obj, datetime.timedelta):
        res = obj.total_seconds()

    elif isinstance(obj, enum.Enum):
        res = _to_serializable(obj.value)

    else:
        res = str(obj)

    return res


def _serialize_to_str(value: Any) -> str:
    """Get input attributes

    Args:
        value (`Any`):
            The input value

    Returns:
        `str`:
            JSON serialized string of the input value
    """
    try:
        return json.dumps(value, ensure_ascii=False)

    except TypeError:
        return json.dumps(
            _to_serializable(value),
            ensure_ascii=False,
        )



# -*- coding: utf-8 -*-
"""The metaclass for agents in agentscope."""
import inspect
from copy import deepcopy
from functools import wraps
from typing import (
    Any,
    Dict,
    TYPE_CHECKING,
    Callable,
)

from .._utils._common import _execute_async_or_sync_func

if TYPE_CHECKING:
    from ._agent_base import AgentBase
else:
    AgentBase = "AgentBase"


def _normalize_to_kwargs(
    func: Callable,
    self: Any,
    *args: Any,
    **kwargs: Any,
) -> dict:
    """Normalize the provided positional and keyword arguments into a
    keyword arguments dictionary that matches the function signature."""
    sig = inspect.signature(func)
    try:
        # Bind the provided arguments to the function signature
        bound = sig.bind(self, *args, **kwargs)
        # Apply the default values for parameters
        bound.apply_defaults()

        # Return the arguments in a dictionary format
        res = dict(bound.arguments)
        res.pop("self")
        return res

    except TypeError as e:
        # If failed to bind, we raise a TypeError with more context
        param_names = list(sig.parameters.keys())
        provided_args = len(args)
        provided_kwargs = list(kwargs.keys())

        raise TypeError(
            f"Failed to bind parameters for function '{func.__name__}': {e}\n"
            f"Expected parameters: {param_names}\n"
            f"Provided {provided_args} positional args and kwargs: "
            f"{provided_kwargs}",
        ) from e


def _wrap_with_hooks(
    original_func: Callable,
) -> Callable:
    """A decorator to wrap the original async function with pre- and post-hooks

    Args:
        original_func (`Callable`):
            The original async function to be wrapped with hooks.
    """
    func_name = original_func.__name__.replace("_", "")

    @wraps(original_func)
    async def async_wrapper(
        self: AgentBase,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """The wrapped function, which call the pre- and post-hooks before and
        after the original function."""

        # Unify all positional and keyword arguments into a keyword arguments
        normalized_kwargs = _normalize_to_kwargs(
            original_func,
            self,
            *args,
            **kwargs,
        )

        current_normalized_kwargs = normalized_kwargs
        assert (
            hasattr(self, f"_instance_pre_{func_name}_hooks")
            and hasattr(self, f"_instance_post_{func_name}_hooks")
            and hasattr(self.__class__, f"_class_pre_{func_name}_hooks")
            and hasattr(self.__class__, f"_class_post_{func_name}_hooks")
        ), f"Hooks for {func_name} not found in {self.__class__.__name__}"

        # pre-hooks
        pre_hooks = list(
            getattr(self, f"_instance_pre_{func_name}_hooks").values(),
        ) + list(
            getattr(self, f"_class_pre_{func_name}_hooks").values(),
        )
        for pre_hook in pre_hooks:
            modified_keywords = await _execute_async_or_sync_func(
                pre_hook,
                self,
                deepcopy(current_normalized_kwargs),
            )
            if modified_keywords is not None:
                assert isinstance(modified_keywords, dict), (
                    f"Pre-hook must return a dict of keyword arguments, rather"
                    f" than {type(modified_keywords)} from hook "
                    f"{pre_hook.__name__}"
                )
                current_normalized_kwargs = modified_keywords

        # original function
        # handle positional and keyword arguments specifically
        args = current_normalized_kwargs.get("args", [])
        kwargs = current_normalized_kwargs.get("kwargs", {})
        others = {
            k: v
            for k, v in current_normalized_kwargs.items()
            if k not in ["args", "kwargs"]
        }
        current_output = await original_func(
            self,
            *args,
            **others,
            **kwargs,
        )

        # post_hooks
        post_hooks = list(
            getattr(self, f"_instance_post_{func_name}_hooks").values(),
        ) + list(
            getattr(self, f"_class_post_{func_name}_hooks").values(),
        )
        for post_hook in post_hooks:
            modified_output = await _execute_async_or_sync_func(
                post_hook,
                self,
                deepcopy(current_normalized_kwargs),
                deepcopy(current_output),
            )
            if modified_output is not None:
                current_output = modified_output
        return current_output

    return async_wrapper


class _AgentMeta(type):
    """The agent metaclass that wraps the agent's reply, observe and print
    functions with pre- and post-hooks."""

    def __new__(mcs, name: Any, bases: Any, attrs: Dict) -> Any:
        """Wrap the agent's functions with hooks."""

        for func_name in [
            "reply",
            "print",
            "observe",
        ]:
            if func_name in attrs:
                attrs[func_name] = _wrap_with_hooks(attrs[func_name])

        return super().__new__(mcs, name, bases, attrs)


class _ReActAgentMeta(_AgentMeta):
    """The ReAct metaclass that adds pre- and post-hooks for the _reasoning
    and _acting functions."""

    def __new__(mcs, name: Any, bases: Any, attrs: Dict) -> Any:
        """Wrap the ReAct agent's _reasoning and _acting functions with
        hooks."""

        for func_name in [
            "_reasoning",
            "_acting",
        ]:
            if func_name in attrs:
                attrs[func_name] = _wrap_with_hooks(attrs[func_name])

        return super().__new__(mcs, name, bases, attrs)



# -*- coding: utf-8 -*-
"""The embedding response class."""
from dataclasses import dataclass, field
from typing import Literal, List

from ._embedding_usage import EmbeddingUsage
from .._utils._common import _get_timestamp
from .._utils._mixin import DictMixin
from ..types import Embedding


@dataclass
class EmbeddingResponse(DictMixin):
    """The embedding response class."""

    embeddings: List[Embedding]
    """The embedding data"""

    id: str = field(default_factory=lambda: _get_timestamp(True))
    """The identity of the embedding response"""

    created_at: str = field(default_factory=_get_timestamp)
    """The timestamp of the embedding response creation"""

    type: Literal["embedding"] = field(default_factory=lambda: "embedding")
    """The type of the response, must be `embedding`."""

    usage: EmbeddingUsage | None = field(default_factory=lambda: None)
    """The usage of the embedding model API invocation, if available."""

    source: Literal["cache", "api"] = field(default_factory=lambda: "api")
    """If the response comes from the cache or the API."""



# -*- coding: utf-8 -*-
"""The reader abstraction for retrieval-augmented generation (RAG)."""

from ._reader_base import ReaderBase, Document
from ._text_reader import TextReader
from ._pdf_reader import PDFReader
from ._word_reader import WordReader
from ._image_reader import ImageReader


__all__ = [
    "Document",
    "ReaderBase",
    "TextReader",
    "PDFReader",
    "WordReader",
    "ImageReader",
]



# -*- coding: utf-8 -*-
"""The evaluation module in AgentScope."""

from ._evaluator import (
    EvaluatorBase,
    RayEvaluator,
    GeneralEvaluator,
)
from ._metric_base import (
    MetricBase,
    MetricResult,
    MetricType,
)
from ._task import Task
from ._solution import SolutionOutput
from ._benchmark_base import BenchmarkBase
from ._evaluator_storage import (
    EvaluatorStorageBase,
    FileEvaluatorStorage,
)
from ._ace_benchmark import (
    ACEBenchmark,
    ACEAccuracy,
    ACEProcessAccuracy,
    ACEPhone,
)

__all__ = [
    "BenchmarkBase",
    "EvaluatorBase",
    "RayEvaluator",
    "GeneralEvaluator",
    "MetricBase",
    "MetricResult",
    "MetricType",
    "EvaluatorStorageBase",
    "FileEvaluatorStorage",
    "Task",
    "SolutionOutput",
    "ACEBenchmark",
    "ACEAccuracy",
    "ACEProcessAccuracy",
    "ACEPhone",
]



# -*- coding: utf-8 -*-
"""Solution class for evaluation tasks."""

from dataclasses import dataclass, field
from typing import Any

from ..message import (
    ToolResultBlock,
    ToolUseBlock,
    TextBlock,
)
from ..types._json import JSONSerializableObject
from .._utils._mixin import DictMixin


@dataclass
class SolutionOutput(DictMixin):
    """The output of a solution in evaluation task"""

    success: bool
    """Indicates whether the solution is executed successfully. When the
    solution raise exception, this should be set to False."""
    output: JSONSerializableObject
    """The final output of the solution."""
    trajectory: list[ToolUseBlock | ToolResultBlock | TextBlock]
    """The tool calls and results trajectory"""
    meta: dict[str, Any] | None = field(default_factory=lambda: None)
    """Additional metadata for the solution"""

    def __getstate__(self) -> dict[str, Any]:
        """Custom pickling to handle dataclass + DictMixin inheritance."""
        return self.__dict__.copy()

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Custom unpickling to handle dataclass + DictMixin inheritance."""
        self.__dict__.update(state)



# -*- coding: utf-8 -*-
"""The token base class in agentscope."""
from abc import abstractmethod
from typing import Any


class TokenCounterBase:
    """The base class for token counting."""

    @abstractmethod
    async def count(
        self,
        messages: list[dict],
        **kwargs: Any,
    ) -> int:
        """Count the number of tokens by the given model and messages."""



# -*- coding: utf-8 -*-
"""The JSON session class."""
import json
import os

from ._session_base import SessionBase
from .._logging import logger
from ..module import StateModule


class JSONSession(SessionBase):
    """The JSON session class."""

    def __init__(
        self,
        session_id: str | None = None,
        save_dir: str = "./",
    ) -> None:
        """Initialize the JSON session class.

        Args:
            session_id (`str`):
                The session id, deprecated and move to the `save_session_state`
                and `load_session_state` methods to support different session
                ids.
            save_dir (`str`, defaults to `"./"):
                The directory to save the session state.
        """
        self.save_dir = save_dir

        if session_id is not None:
            logger.warning(
                "The `session_id` argument in the JSONSession constructor is "
                "deprecated and will be removed in future versions. Please "
                "pass the `session_id` to the `save_session_state` and "
                "`load_session_state` methods instead.",
            )

    def _get_save_path(self, session_id: str) -> str:
        """The path to save the session state.

        Args:
            session_id (`str`):
                The session id.

        Returns:
            `str`:
                The path to save the session state.
        """
        os.makedirs(self.save_dir, exist_ok=True)
        return os.path.join(self.save_dir, f"{session_id}.json")

    async def save_session_state(
        self,
        session_id: str,
        **state_modules_mapping: StateModule,
    ) -> None:
        """Load the state dictionary from a JSON file.

        Args:
            session_id (`str`):
                The session id.
            **state_modules_mapping (`dict[str, StateModule]`):
                A dictionary mapping of state module names to their instances.
        """
        state_dicts = {
            name: state_module.state_dict()
            for name, state_module in state_modules_mapping.items()
        }
        with open(
            self._get_save_path(session_id),
            "w",
            encoding="utf-8",
        ) as file:
            json.dump(state_dicts, file, ensure_ascii=False)

    async def load_session_state(
        self,
        session_id: str,
        allow_not_exist: bool = True,
        **state_modules_mapping: StateModule,
    ) -> None:
        """Get the state dictionary to be saved to a JSON file.

        Args:
            session_id (`str`):
                The session id.
            allow_not_exist (`bool`, defaults to `True`):
                Whether to allow the session to not exist. If `False`, raises
                an error if the session does not exist.
            state_modules_mapping (`list[StateModule]`):
                The list of state modules to be loaded.
        """
        session_save_path = self._get_save_path(session_id)
        if os.path.exists(session_save_path):
            with open(session_save_path, "r", encoding="utf-8") as file:
                states = json.load(file)

            for name, state_module in state_modules_mapping.items():
                if name in states:
                    state_module.load_state_dict(states[name])
            logger.info(
                "Load session state from %s successfully.",
                session_save_path,
            )

        elif allow_not_exist:
            logger.info(
                "Session file %s does not exist. Skip loading session state.",
                session_save_path,
            )

        else:
            raise ValueError(
                f"Failed to load session state for file {session_save_path} "
                "does not exist.",
            )



# -*- coding: utf-8 -*-
# pylint: disable=not-an-iterable
# mypy: disable-error-code="list-item"
"""ReAct agent class in agentscope."""
import asyncio
from typing import Type, Any, AsyncGenerator, Literal

import shortuuid
from pydantic import BaseModel, ValidationError, Field

from ._react_agent_base import ReActAgentBase
from .._logging import logger
from ..formatter import FormatterBase
from ..memory import MemoryBase, LongTermMemoryBase, InMemoryMemory
from ..message import Msg, ToolUseBlock, ToolResultBlock, TextBlock
from ..model import ChatModelBase
from ..rag import KnowledgeBase, Document
from ..plan import PlanNotebook
from ..tool import Toolkit, ToolResponse
from ..tracing import trace_reply


class _QueryRewriteModel(BaseModel):
    """The structured model used for query rewriting."""

    rewritten_query: str = Field(
        description=(
            "The rewritten query, which should be specific and concise. "
        ),
    )


def finish_function_pre_print_hook(
    self: "ReActAgent",
    kwargs: dict[str, Any],
) -> dict[str, Any] | None:
    """A pre-speak hook function that check if finish_function is called. If
    so, it will wrap the response argument into a message and return it to
    replace the original message. By this way, the calling of the finish
    function will be displayed as a text reply instead of a tool call."""

    msg = kwargs["msg"]

    if isinstance(msg.content, str):
        return None

    if isinstance(msg.content, list):
        for i, block in enumerate(msg.content):
            if (
                block["type"] == "tool_use"
                and block["name"] == self.finish_function_name
            ):
                # Convert the response argument into a text block for
                # displaying
                try:
                    msg.content[i] = TextBlock(
                        type="text",
                        text=block["input"].get("response", ""),
                    )
                    return kwargs
                except Exception:
                    print("Error in block input", block["input"])

    return None


class ReActAgent(ReActAgentBase):
    """A ReAct agent implementation in AgentScope, which supports

    - Realtime steering
    - API-based (parallel) tool calling
    - Hooks around reasoning, acting, reply, observe and print functions
    - Structured output generation
    """

    finish_function_name: str = "generate_response"
    """The function name used to finish replying and return a response to
    the user."""

    def __init__(
        self,
        name: str,
        sys_prompt: str,
        model: ChatModelBase,
        formatter: FormatterBase,
        toolkit: Toolkit | None = None,
        memory: MemoryBase | None = None,
        long_term_memory: LongTermMemoryBase | None = None,
        long_term_memory_mode: Literal[
            "agent_control",
            "static_control",
            "both",
        ] = "both",
        enable_meta_tool: bool = False,
        parallel_tool_calls: bool = False,
        knowledge: KnowledgeBase | list[KnowledgeBase] | None = None,
        enable_rewrite_query: bool = True,
        plan_notebook: PlanNotebook | None = None,
        print_hint_msg: bool = False,
        max_iters: int = 10,
    ) -> None:
        """Initialize the ReAct agent

        Args:
            name (`str`):
                The name of the agent.
            sys_prompt (`str`):
                The system prompt of the agent.
            model (`ChatModelBase`):
                The chat model used by the agent.
            formatter (`FormatterBase`):
                The formatter used to format the messages into the required
                format of the model API provider.
            toolkit (`Toolkit | None`, optional):
                A `Toolkit` object that contains the tool functions. If not
                provided, a default empty `Toolkit` will be created.
            memory (`MemoryBase | None`, optional):
                The memory used to store the dialogue history. If not provided,
                a default `InMemoryMemory` will be created, which stores
                messages in a list in memory.
            long_term_memory (`LongTermMemoryBase | None`, optional):
                The optional long-term memory, which will provide two tool
                functions: `retrieve_from_memory` and `record_to_memory`, and
                will attach the retrieved information to the system prompt
                before each reply.
            enable_meta_tool (`bool`, defaults to `False`):
                If `True`, a meta tool function `reset_equipped_tools` will be
                added to the toolkit, which allows the agent to manage its
                equipped tools dynamically.
            long_term_memory_mode (`Literal['agent_control', 'static_control',\
              'both']`, defaults to `both`):
                The mode of the long-term memory. If `agent_control`, two
                tool functions `retrieve_from_memory` and `record_to_memory`
                will be registered in the toolkit to allow the agent to
                manage the long-term memory. If `static_control`, retrieving
                and recording will happen in the beginning and end of
                each reply respectively.
            parallel_tool_calls (`bool`, defaults to `False`):
                When LLM generates multiple tool calls, whether to execute
                them in parallel.
            knowledge (`KnowledgeBase | list[KnowledgeBase] | None`, optional):
                The knowledge object(s) used by the agent to retrieve
                relevant documents at the beginning of each reply.
            enable_rewrite_query (`bool`, defaults to `True`):
                Whether ask the agent to rewrite the user input query before
                retrieving from the knowledge base(s), e.g. rewrite "Who am I"
                to "{user's name}" to get more relevant documents. Only works
                when the knowledge base(s) is provided.
            plan_notebook (`PlanNotebook | None`, optional):
                The plan notebook instance, allow the agent to finish the
                complex task by decomposing it into a sequence of subtasks.
            print_hint_msg (`bool`, defaults to `False`):
                Whether to print the hint messages, including the reasoning
                hint from the plan notebook, the retrieved information from
                the long-term memory and knowledge base(s).
            max_iters (`int`, defaults to `10`):
                The maximum number of iterations of the reasoning-acting loops.
        """
        super().__init__()

        assert long_term_memory_mode in [
            "agent_control",
            "static_control",
            "both",
        ]

        # Static variables in the agent
        self.name = name
        self._sys_prompt = sys_prompt
        self.max_iters = max_iters
        self.model = model
        self.formatter = formatter

        # -------------- Memory management --------------
        # Record the dialogue history in the memory
        self.memory = memory or InMemoryMemory()
        # If provide the long-term memory, it will be used to retrieve info
        # in the beginning of each reply, and the result will be added to the
        # system prompt
        self.long_term_memory = long_term_memory

        # The long-term memory mode
        self._static_control = long_term_memory and long_term_memory_mode in [
            "static_control",
            "both",
        ]
        self._agent_control = long_term_memory and long_term_memory_mode in [
            "agent_control",
            "both",
        ]

        # -------------- Tool management --------------
        # If None, a default Toolkit will be created
        self.toolkit = toolkit or Toolkit()
        self.toolkit.register_tool_function(
            getattr(self, self.finish_function_name),
        )
        if self._agent_control:
            # Adding two tool functions into the toolkit to allow self-control
            self.toolkit.register_tool_function(
                long_term_memory.record_to_memory,
            )
            self.toolkit.register_tool_function(
                long_term_memory.retrieve_from_memory,
            )
        # Add a meta tool function to allow agent-controlled tool management
        if enable_meta_tool or plan_notebook:
            self.toolkit.register_tool_function(
                self.toolkit.reset_equipped_tools,
            )

        self.parallel_tool_calls = parallel_tool_calls

        # -------------- RAG management --------------
        # The knowledge base(s) used by the agent
        if isinstance(knowledge, KnowledgeBase):
            knowledge = [knowledge]
        self.knowledge: list[KnowledgeBase] = knowledge or []
        self.enable_rewrite_query = enable_rewrite_query

        # -------------- Plan management --------------
        # Equipped the plan-related tools provided by the plan notebook as
        # a tool group named "plan_related". So that the agent can activate
        # the plan tools by the meta tool function
        self.plan_notebook = None
        if plan_notebook:
            self.plan_notebook = plan_notebook
            # When enable_meta_tool is True, plan tools are in plan_related
            # group and active by agent.
            # Otherwise, plan tools in bassic group and always active.
            if enable_meta_tool:
                self.toolkit.create_tool_group(
                    "plan_related",
                    description=self.plan_notebook.description,
                )
                for tool in plan_notebook.list_tools():
                    self.toolkit.register_tool_function(
                        tool,
                        group_name="plan_related",
                    )
            else:
                for tool in plan_notebook.list_tools():
                    self.toolkit.register_tool_function(
                        tool,
                    )

        # If print the reasoning hint messages
        self.print_hint_msg = print_hint_msg

        # The maximum number of iterations of the reasoning-acting loops
        self.max_iters = max_iters

        # The hint messages that will be attached to the prompt to guide the
        # agent's behavior before each reasoning step, and cleared after
        # each reasoning step, meaning the hint messages is one-time use only.
        # We use an InMemoryMemory instance to store the hint messages
        self._reasoning_hint_msgs = InMemoryMemory()

        # Variables to record the intermediate state

        # If required structured output model is provided
        self._required_structured_model: Type[BaseModel] | None = None

        # -------------- State registration and hooks --------------
        # Register the status variables
        self.register_state("name")
        self.register_state("_sys_prompt")

        self.register_instance_hook(
            "pre_print",
            "finish_function_pre_print_hook",
            finish_function_pre_print_hook,
        )

    @property
    def sys_prompt(self) -> str:
        """The dynamic system prompt of the agent."""
        return self._sys_prompt

    @trace_reply
    async def reply(
        self,
        msg: Msg | list[Msg] | None = None,
        structured_model: Type[BaseModel] | None = None,
    ) -> Msg:
        """Generate a reply based on the current state and input arguments.

        Args:
            msg (`Msg | list[Msg] | None`, optional):
                The input message(s) to the agent.
            structured_model (`Type[BaseModel] | None`, optional):
                The required structured output model. If provided, the agent
                is expected to generate structured output in the `metadata`
                field of the output message.

        Returns:
            `Msg`:
                The output message generated by the agent.
        """
        # Record the input message(s) in the memory
        await self.memory.add(msg)

        # Retrieve relevant records from the long-term memory if activated
        await self._retrieve_from_long_term_memory(msg)
        # Retrieve relevant documents from the knowledge base(s) if any
        await self._retrieve_from_knowledge(msg)

        self._required_structured_model = structured_model
        # Record structured output model if provided
        if structured_model:
            self.toolkit.set_extended_model(
                self.finish_function_name,
                structured_model,
            )

        # The reasoning-acting loop
        reply_msg = None
        for _ in range(self.max_iters):
            msg_reasoning = await self._reasoning()

            futures = [
                self._acting(tool_call)
                for tool_call in msg_reasoning.get_content_blocks(
                    "tool_use",
                )
            ]

            # Parallel tool calls or not
            if self.parallel_tool_calls:
                acting_responses = await asyncio.gather(*futures)

            else:
                # Sequential tool calls
                acting_responses = [await _ for _ in futures]

            # Find the first non-None replying message from the acting
            for acting_msg in acting_responses:
                reply_msg = reply_msg or acting_msg

            if reply_msg:
                break

        # When the maximum iterations are reached
        if reply_msg is None:
            reply_msg = await self._summarizing()

        # Post-process the memory, long-term memory
        if self._static_control:
            await self.long_term_memory.record(
                [
                    *([*msg] if isinstance(msg, list) else [msg]),
                    *await self.memory.get_memory(),
                    reply_msg,
                ],
            )

        await self.memory.add(reply_msg)
        return reply_msg

    async def _reasoning(
        self,
    ) -> Msg:
        """Perform the reasoning process."""
        if self.plan_notebook:
            # Insert the reasoning hint from the plan notebook
            hint_msg = await self.plan_notebook.get_current_hint()
            if self.print_hint_msg and hint_msg:
                await self.print(hint_msg)
            await self._reasoning_hint_msgs.add(hint_msg)

        # Convert Msg objects into the required format of the model API
        prompt = await self.formatter.format(
            msgs=[
                Msg("system", self.sys_prompt, "system"),
                *await self.memory.get_memory(),
                # The hint messages to guide the agent's behavior, maybe empty
                *await self._reasoning_hint_msgs.get_memory(),
            ],
        )
        # Clear the hint messages after use
        await self._reasoning_hint_msgs.clear()

        res = await self.model(
            prompt,
            tools=self.toolkit.get_json_schemas(),
        )

        # handle output from the model
        interrupted_by_user = False
        msg = None
        try:
            if self.model.stream:
                msg = Msg(self.name, [], "assistant")
                async for content_chunk in res:
                    msg.content = content_chunk.content
                    await self.print(msg, False)
                await self.print(msg, True)

                # Add a tiny sleep to yield the last message object in the
                # message queue
                await asyncio.sleep(0.001)

            else:
                msg = Msg(self.name, list(res.content), "assistant")
                await self.print(msg, True)

        except asyncio.CancelledError as e:
            interrupted_by_user = True
            raise e from None

        finally:
            if msg and not msg.has_content_blocks("tool_use"):
                # Turn plain text response into a tool call of the finish
                # function
                msg = Msg.from_dict(msg.to_dict())
                msg.content = [
                    ToolUseBlock(
                        id=shortuuid.uuid(),
                        type="tool_use",
                        name=self.finish_function_name,
                        input={"response": msg.get_text_content()},
                    ),
                ]

            # None will be ignored by the memory
            await self.memory.add(msg)

            # Post-process for user interruption
            if interrupted_by_user and msg:
                # Fake tool results
                tool_use_blocks: list = msg.get_content_blocks(
                    "tool_use",
                )
                for tool_call in tool_use_blocks:
                    msg_res = Msg(
                        "system",
                        [
                            ToolResultBlock(
                                type="tool_result",
                                id=tool_call["id"],
                                name=tool_call["name"],
                                output="The tool call has been interrupted "
                                "by the user.",
                            ),
                        ],
                        "system",
                    )
                    await self.memory.add(msg_res)
                    await self.print(msg_res, True)
        return msg

    async def _acting(self, tool_call: ToolUseBlock) -> Msg | None:
        """Perform the acting process.

        Args:
            tool_call (`ToolUseBlock`):
                The tool use block to be executed.

        Returns:
            `Union[Msg, None]`:
                Return a message to the user if the `finish_function` is
                called, otherwise return `None`.
        """

        tool_res_msg = Msg(
            "system",
            [
                ToolResultBlock(
                    type="tool_result",
                    id=tool_call["id"],
                    name=tool_call["name"],
                    output=[],
                ),
            ],
            "system",
        )
        try:
            # Execute the tool call
            tool_res = await self.toolkit.call_tool_function(tool_call)

            response_msg = None
            # Async generator handling
            async for chunk in tool_res:
                # Turn into a tool result block
                tool_res_msg.content[0][  # type: ignore[index]
                    "output"
                ] = chunk.content

                # Skip the printing of the finish function call
                if (
                    tool_call["name"] != self.finish_function_name
                    or tool_call["name"] == self.finish_function_name
                    and (
                        chunk.metadata is None
                        or not chunk.metadata.get("success")
                    )
                ):
                    await self.print(tool_res_msg, chunk.is_last)

                # Raise the CancelledError to handle the interruption in the
                # handle_interrupt function
                if chunk.is_interrupted:
                    raise asyncio.CancelledError()

                # Return message if generate_response is called successfully
                if (
                    tool_call["name"] == self.finish_function_name
                    and chunk.metadata
                    and chunk.metadata.get(
                        "success",
                        True,
                    )
                ):
                    response_msg = chunk.metadata.get("response_msg")

            return response_msg

        finally:
            # Record the tool result message in the memory
            await self.memory.add(tool_res_msg)

    async def observe(self, msg: Msg | list[Msg] | None) -> None:
        """Receive observing message(s) without generating a reply.

        Args:
            msg (`Msg | list[Msg] | None`):
                The message or messages to be observed.
        """
        await self.memory.add(msg)

    async def _summarizing(self) -> Msg:
        """Generate a response when the agent fails to solve the problem in
        the maximum iterations."""
        hint_msg = Msg(
            "user",
            "You have failed to generate response within the maximum "
            "iterations. Now respond directly by summarizing the current "
            "situation.",
            role="user",
        )

        # Generate a reply by summarizing the current situation
        prompt = await self.formatter.format(
            [
                Msg("system", self.sys_prompt, "system"),
                *await self.memory.get_memory(),
                hint_msg,
            ],
        )
        # TODO: handle the structured output here, maybe force calling the
        #  finish_function here
        res = await self.model(prompt)

        res_msg = Msg(self.name, [], "assistant")
        if isinstance(res, AsyncGenerator):
            async for chunk in res:
                res_msg.content = chunk.content
                await self.print(res_msg, False)
            await self.print(res_msg, True)

        else:
            res_msg.content = res.content
            await self.print(res_msg, True)

        return res_msg

    async def handle_interrupt(
        self,
        _msg: Msg | list[Msg] | None = None,
    ) -> Msg:
        """The post-processing logic when the reply is interrupted by the
        user or something else."""

        response_msg = Msg(
            self.name,
            "I noticed that you have interrupted me. What can I "
            "do for you?",
            "assistant",
            metadata={
                # Expose this field to indicate the interruption
                "is_interrupted": True,
            },
        )

        await self.print(response_msg, True)
        await self.memory.add(response_msg)
        return response_msg

    def generate_response(
        self,
        response: str,
        **kwargs: Any,
    ) -> ToolResponse:
        """Generate a response. Note only the input argument `response` is
        visible to the others, you should include all the necessary
        information in the `response` argument.

        Args:
            response (`str`):
                Your response to the user.
        """
        response_msg = Msg(
            self.name,
            response,
            "assistant",
        )

        # Prepare structured output
        if self._required_structured_model:
            try:
                # Use the metadata field of the message to store the
                # structured output
                response_msg.metadata = (
                    self._required_structured_model.model_validate(
                        kwargs,
                    ).model_dump()
                )

            except ValidationError as e:
                return ToolResponse(
                    content=[
                        TextBlock(
                            type="text",
                            text=f"Arguments Validation Error: {e}",
                        ),
                    ],
                    metadata={
                        "success": False,
                        "response_msg": None,
                    },
                )

        return ToolResponse(
            content=[
                TextBlock(
                    type="text",
                    text="Successfully generated response.",
                ),
            ],
            metadata={
                "success": True,
                "response_msg": response_msg,
            },
            is_last=True,
        )

    async def _retrieve_from_long_term_memory(
        self,
        msg: Msg | list[Msg] | None,
    ) -> None:
        """Insert the retrieved information from the long-term memory into
        the short-term memory as a Msg object.

        Args:
            msg (`Msg | list[Msg] | None`):
                The input message to the agent.
        """
        if self._static_control and msg:
            # Retrieve information from the long-term memory if available
            retrieved_info = await self.long_term_memory.retrieve(msg)
            if retrieved_info:
                retrieved_msg = Msg(
                    name="long_term_memory",
                    content="<long_term_memory>The content below are "
                    "retrieved from long-term memory, which maybe "
                    f"useful:\n{retrieved_info}</long_term_memory>",
                    role="user",
                )
                if self.print_hint_msg:
                    await self.print(retrieved_msg, True)
                await self.memory.add(retrieved_msg)

    async def _retrieve_from_knowledge(
        self,
        msg: Msg | list[Msg] | None,
    ) -> None:
        """Insert the retrieved documents from the RAG knowledge base(s) if
        available.

        Args:
            msg (`Msg | list[Msg] | None`):
                The input message to the agent.
        """
        if self.knowledge and msg:
            # Prepare the user input query
            query = None
            if isinstance(msg, Msg):
                query = msg.get_text_content()
            elif isinstance(msg, list):
                query = "\n".join(_.get_text_content() for _ in msg)

            # Skip if the query is empty
            if not query:
                return

            # Rewrite the query by the LLM if enabled
            if self.enable_rewrite_query:
                try:
                    rewrite_prompt = await self.formatter.format(
                        msgs=[
                            Msg("system", self.sys_prompt, "system"),
                            *await self.memory.get_memory(),
                            Msg(
                                "user",
                                "<system-hint>Now you need to rewrite "
                                "the above user query to be more specific and "
                                "concise for knowledge retrieval. For "
                                "example, rewrite the query 'what happened "
                                "last day' to 'what happened on 2023-10-01' "
                                "(assuming today is 2023-10-02)."
                                "</system-hint>",
                                "user",
                            ),
                        ],
                    )
                    stream_tmp = self.model.stream
                    self.model.stream = False
                    res = await self.model(
                        rewrite_prompt,
                        structured_model=_QueryRewriteModel,
                    )
                    self.model.stream = stream_tmp
                    if res.metadata and res.metadata.get("rewritten_query"):
                        query = res.metadata["rewritten_query"]

                except Exception as e:
                    logger.warning(
                        "Skipping the query rewriting due to error: %s",
                        str(e),
                    )

            docs: list[Document] = []
            for kb in self.knowledge:
                # retrieve the user input query
                docs.extend(
                    await kb.retrieve(query=query),
                )
            if docs:
                # Rerank by the relevance score
                docs = sorted(
                    docs,
                    key=lambda doc: doc.score or 0.0,
                    reverse=True,
                )
                # Prepare the retrieved knowledge string
                retrieved_msg = Msg(
                    name="user",
                    content=[
                        TextBlock(
                            type="text",
                            text=(
                                "<retrieved_knowledge>Use the following "
                                "content from the knowledge base(s) if it's "
                                "helpful:\n"
                            ),
                        ),
                        *[_.metadata.content for _ in docs],
                        TextBlock(
                            type="text",
                            text="</retrieved_knowledge>",
                        ),
                    ],
                    role="user",
                )
                if self.print_hint_msg:
                    await self.print(retrieved_msg, True)
                await self.memory.add(retrieved_msg)



# -*- coding: utf-8 -*-
"""The vector database store abstraction in AgentScope RAG module."""

from ._store_base import (
    VDBStoreBase,
)
from ._qdrant_store import QdrantStore
from ._milvuslite_store import MilvusLiteStore

__all__ = [
    "VDBStoreBase",
    "QdrantStore",
    "MilvusLiteStore",
]



# -*- coding: utf-8 -*-
"""The food platform API in the ACEBench evaluation."""

from ._shared_state import SharedState


class FoodPlatformApi(SharedState):
    """The food platform Api in the ACEBench evaluation."""

    tool_functions: list[str] = [
        "login_food_platform",
        "view_logged_in_users",
        "check_balance",
        "add_food_delivery_order",
        "get_products",
        "view_orders",
        "search_orders",
    ]

    def __init__(self, shared_state: dict) -> None:
        super().__init__(shared_state)

        # 设置用户和初始金额
        self.users: dict = {
            "Eve": {
                "user_id": "U100",
                "password": "password123",
                "balance": 500.0,
            },
            "Frank": {
                "user_id": "U101",
                "password": "password456",
                "balance": 300.0,
            },
            "Grace": {
                "user_id": "U102",
                "password": "password789",
                "balance": 150.0,
            },
            "Helen": {
                "user_id": "U103",
                "password": "password321",
                "balance": 800.0,
            },
            "Isaac": {
                "user_id": "U104",
                "password": "password654",
                "balance": 400.0,
            },
            "Jack": {
                "user_id": "U105",
                "password": "password654",
                "balance": 120.0,
            },
        }

        # 设置六个商家及其菜单
        self.merchant_list: dict[str, dict] = {
            "达美乐": {
                "merchant_id": "M100",
                "service_type": "Pizza",
                "menu": [
                    {"product": "玛格丽特披萨", "price": 68.0},
                    {"product": "超级至尊披萨", "price": 88.0},
                ],
            },
            "米村拌饭": {
                "merchant_id": "M101",
                "service_type": "Bibimbap",
                "menu": [
                    {"product": "石锅拌饭", "price": 35.0},
                    {"product": "韩式牛肉拌饭", "price": 45.0},
                ],
            },
            "海底捞": {
                "merchant_id": "M102",
                "service_type": "Hotpot",
                "menu": [
                    {"product": "牛肉卷", "price": 68.0},
                    {"product": "海鲜拼盘", "price": 88.0},
                ],
            },
            "喜茶": {
                "merchant_id": "M103",
                "service_type": "Milk Tea",
                "menu": [
                    {"product": "芝士奶茶", "price": 25.0},
                    {"product": "四季春奶茶", "price": 22.0},
                ],
            },
            "盒马生鲜": {
                "merchant_id": "M104",
                "service_type": "Fresh Grocery",
                "menu": [
                    {"product": "有机蔬菜包", "price": 15.0},
                    {"product": "生鲜大礼包", "price": 99.0},
                ],
            },
            "九田家烤肉": {
                "merchant_id": "M105",
                "service_type": "BBQ",
                "menu": [
                    {"product": "韩式烤牛肉", "price": 128.0},
                    {"product": "烤五花肉", "price": 78.0},
                ],
            },
        }

        # 设置已登录用户列表
        self.logged_in_users: list[str] = []
        # 订单列表
        self.orders: list = []

    def get_state_dict(self) -> dict:
        """Get the current state dict of the FoodPlatformApi."""
        return {
            "FoodPlatform": {
                "logged_in_users": self.logged_in_users,
                "orders": self.orders,
                "users": self.users,
            },
        }

    def login_food_platform(
        self,
        username: str,
        password: str,
    ) -> dict[str, bool | str]:
        """使用用户名和密码登录外卖平台。

        Args:
            username (`str`):
                用户的用户名。
            password (`str`):
                用户的密码。
        """
        if not self.wifi:
            return {"status": False, "message": "wifi未打开，无法登录"}
        if username not in self.users:
            return {"status": False, "message": "用户不存在"}
        if self.users[username]["password"] != password:
            return {"status": False, "message": "密码错误"}

        # 检查是否已经有用户登录
        if username in self.logged_in_users:
            return {"status": False, "message": f"{username} 已经登录"}

        # 记录已登录用户
        self.logged_in_users.append(username)
        return {"status": True, "message": f"用户{username}登陆成功！"}

    def view_logged_in_users(self) -> dict:
        """查看当前所有登录的用户。"""
        if not self.logged_in_users:
            return {
                "status": False,
                "message": "当前没有登录food platform",
            }

        return {"status": True, "logged_in_users": self.logged_in_users}

    def check_balance(self, user_name: str) -> float:
        """查询指定用户的余额。

        Args:
            user_name (`str`):
                用户的用户名。
        """
        if user_name in self.users:
            return self.users[user_name]["balance"]
        else:
            return 0.0

    def add_food_delivery_order(
        self,
        username: str,
        merchant_name: str,
        items: list[dict[str, str | int]],
    ) -> dict[str, bool | str]:
        """订外卖

        Args:
            username (`str`):
                下订单的用户姓名。
            merchant_name (`str`):
                下订单的商家名称。
            items (`list[dict[str, str | int]]`):
                订单中商品的列表，每个商品包含名称和数量。
        """
        if username not in self.logged_in_users:
            return {
                "status": False,
                "message": f"用户 {username} 未登录food platform",
            }

        if merchant_name not in self.merchant_list:
            return {"status": False, "message": "商家不存在"}

        total_price = 0.0
        order_items = []

        for item in items:
            product_name = item.get("product")
            quantity = item.get("quantity", 1)

            if not isinstance(quantity, int) or quantity <= 0:
                return {
                    "status": False,
                    "message": f"无效的数量 {quantity} 对于商品 {product_name}",
                }

            # 查找商品价格
            product_found = False
            for product in self.merchant_list[merchant_name]["menu"]:
                if product["product"] == product_name:
                    total_price += product["price"] * quantity
                    order_items.append(
                        {
                            "product": product_name,
                            "quantity": quantity,
                            "price_per_unit": product["price"],
                        },
                    )
                    product_found = True
                    break
            if not product_found:
                return {
                    "status": False,
                    "message": f"商品 {product_name} 不存在于 "
                    f"{merchant_name} 的菜单中",
                }

        # 检查余额是否足够
        if total_price >= self.users[username]["balance"]:
            return {"status": False, "message": "余额不足，无法下单"}

        # 扣除余额并创建订单
        self.users[username]["balance"] -= total_price
        order = {
            "user_name": username,
            "merchant_name": merchant_name,
            "items": order_items,
            "total_price": total_price,
        }
        self.orders.append(order)
        return {
            "status": True,
            "message": f"外卖订单成功下单给 {merchant_name}，" f"总金额为 {total_price} 元",
        }

    def get_products(
        self,
        merchant_name: str,
    ) -> list[dict[str, str | float]] | dict[str, bool | str]:
        """获取特定商家的商品列表。

        Args:
            merchant_name (`str`):
                要获取商品的商家名称。
        """
        merchant = self.merchant_list.get(merchant_name)
        if merchant:
            return merchant["menu"]
        else:
            return {
                "status": False,
                "message": f"商家 '{merchant_name}' 不存在",
            }

    def view_orders(
        self,
        user_name: str,
    ) -> dict[str, bool | str | list[dict[str, str | int | float]]]:
        """查看用户的所有订单"""
        user_orders = [
            order for order in self.orders if order["user_name"] == user_name
        ]

        if not user_orders:
            return {"status": False, "message": "用户没有订单记录"}

        return {"status": True, "orders": user_orders}

    def search_orders(
        self,
        keyword: str,
    ) -> dict[str, bool | str | list[dict[str, str | float]]]:
        """根据关键字搜索订单。"""
        matched_orders = [
            order
            for order in self.orders
            if keyword.lower() in order["merchant_name"].lower()
            or any(
                keyword.lower() in item.lower()
                for item in order.get("items", [])
            )
        ]

        if not matched_orders:
            return {"status": False, "message": "没有找到匹配的订单"}

        return {"status": True, "orders": matched_orders}



# -*- coding: utf-8 -*-
"""The chat model base class."""

from abc import abstractmethod
from typing import AsyncGenerator, Any

from ._model_response import ChatResponse

TOOL_CHOICE_MODES = ["auto", "none", "any", "required"]


class ChatModelBase:
    """Base class for chat models."""

    model_name: str
    """The model name"""

    stream: bool
    """Is the model output streaming or not"""

    def __init__(
        self,
        model_name: str,
        stream: bool,
    ) -> None:
        """Initialize the chat model base class.

        Args:
            model_name (`str`):
                The name of the model
            stream (`bool`):
                Whether the model output is streaming or not
        """
        self.model_name = model_name
        self.stream = stream

    @abstractmethod
    async def __call__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> ChatResponse | AsyncGenerator[ChatResponse, None]:
        pass

    def _validate_tool_choice(
        self,
        tool_choice: str,
        tools: list[dict] | None,
    ) -> None:
        """
        Validate tool_choice parameter.

        Args:
            tool_choice (`str`):
                Tool choice mode or function name
            tools (`list[dict] | None`):
                Available tools list
        Raises:
            TypeError: If tool_choice is not string
            ValueError: If tool_choice is invalid
        """
        if not isinstance(tool_choice, str):
            raise TypeError(
                f"tool_choice must be str, got {type(tool_choice)}",
            )
        if tool_choice in TOOL_CHOICE_MODES:
            return

        available_functions = [tool["function"]["name"] for tool in tools]

        if tool_choice not in available_functions:
            all_options = TOOL_CHOICE_MODES + available_functions
            raise ValueError(
                f"Invalid tool_choice '{tool_choice}'. "
                f"Available options: {', '.join(sorted(all_options))}",
            )



# -*- coding: utf-8 -*-
"""The ACE benchmark related implementations in AgentScope."""

from ._ace_benchmark import ACEBenchmark
from ._ace_metric import (
    ACEAccuracy,
    ACEProcessAccuracy,
)
from ._ace_tools_zh import ACEPhone

__all__ = [
    "ACEBenchmark",
    "ACEPhone",
    "ACEAccuracy",
    "ACEProcessAccuracy",
]



# -*- coding: utf-8 -*-
"""The MCP tool function class in AgentScope."""
from contextlib import _AsyncGeneratorContextManager
from typing import Any, Callable

import mcp
from mcp import ClientSession

from ._client_base import MCPClientBase
from .._utils._common import _extract_json_schema_from_mcp_tool
from ..tool import ToolResponse


class MCPToolFunction:
    """An MCP tool function class that can be called directly."""

    name: str
    """The name of the tool function."""

    description: str
    """The description of the tool function."""

    json_schema: dict[str, Any]
    """JSON schema of the tool function"""

    def __init__(
        self,
        mcp_name: str,
        tool: mcp.types.Tool,
        wrap_tool_result: bool,
        client_gen: Callable[..., _AsyncGeneratorContextManager[Any]]
        | None = None,
        session: ClientSession | None = None,
    ) -> None:
        """Initialize the MCP function."""
        self.mcp_name = mcp_name
        self.name = tool.name
        self.description = tool.description
        self.json_schema = _extract_json_schema_from_mcp_tool(tool)
        self.wrap_tool_result = wrap_tool_result

        # Cannot be None at the same time
        if (
            client_gen is None
            and session is None
            or (client_gen is not None and session is not None)
        ):
            raise ValueError(
                "Either client or session must be provided, but not both.",
            )

        self.client_gen = client_gen
        self.session = session

    async def __call__(
        self,
        **kwargs: Any,
    ) -> mcp.types.CallToolResult | ToolResponse:
        """Call the MCP tool function with the given arguments, and return
        the result."""
        if self.client_gen:
            async with self.client_gen() as cli:
                read_stream, write_stream = cli[0], cli[1]
                async with ClientSession(read_stream, write_stream) as session:
                    await session.initialize()
                    res = await session.call_tool(
                        self.name,
                        arguments=kwargs,
                    )

        else:
            res = await self.session.call_tool(
                self.name,
                arguments=kwargs,
            )

        if self.wrap_tool_result:
            as_content = MCPClientBase._convert_mcp_content_to_as_blocks(
                res.content,
            )
            return ToolResponse(
                content=as_content,
                metadata=res.meta,
            )

        return res



# -*- coding: utf-8 -*-
"""The shared state class for ACEBench simulation tools."""


class SharedState:
    """The sharing state class for ACEBench simulation tools."""

    def __init__(self, shared_state: dict) -> None:
        """Initialize the shared state"""
        self._shared_state = shared_state

    @property
    def wifi(self) -> bool:
        """The WI-FI state"""
        return self._shared_state["wifi"]

    @property
    def logged_in(self) -> bool:
        """The logged in state"""
        return self._shared_state["logged_in"]



# -*- coding: utf-8 -*-
"""The base class for plan storage."""
from abc import abstractmethod

from agentscope.module import StateModule
from agentscope.plan._plan_model import Plan


class PlanStorageBase(StateModule):
    """The base class for plan storage."""

    @abstractmethod
    async def add_plan(self, plan: Plan) -> None:
        """Add a plan to the storage."""

    @abstractmethod
    async def delete_plan(self, plan_id: str) -> None:
        """Delete a plan from the storage."""

    @abstractmethod
    async def get_plans(self) -> list[Plan]:
        """Get all plans from the storage."""

    @abstractmethod
    async def get_plan(self, plan_id: str) -> Plan | None:
        """Get a plan by its ID."""



# -*- coding: utf-8 -*-
# flake8: noqa: E501
# pylint: disable=line-too-long
"""The text file tools in agentscope."""
import os

from ._utils import _calculate_view_ranges, _view_text_file
from .._response import ToolResponse
from ...message import TextBlock


async def insert_text_file(
    file_path: str,
    content: str,
    line_number: int,
) -> ToolResponse:
    """Insert the content at the specified line number in a text file.

    Args:
        file_path (`str`):
            The target file path.
        content (`str`):
            The content to be inserted.
        line_number (`int`):
            The line number at which the content should be inserted, starting
            from 1. If exceeds the number of lines in the file, it will be
            appended to the end of the file.

    Returns:
        `ToolResponse`:
            The tool response containing the result of the insertion operation.
    """
    if line_number <= 0:
        return ToolResponse(
            content=[
                TextBlock(
                    type="text",
                    text=f"InvalidArgumentsError: "
                    f"The line number {line_number} is invalid. ",
                ),
            ],
        )

    if not os.path.exists(file_path):
        return ToolResponse(
            content=[
                TextBlock(
                    type="text",
                    text=f"InvalidArgumentsError: The target file "
                    f"{file_path} does not exist. ",
                ),
            ],
        )

    with open(file_path, "r", encoding="utf-8") as file:
        original_lines = file.readlines()

    if line_number == len(original_lines) + 1:
        new_lines = original_lines + ["\n" + content]
    elif line_number < len(original_lines) + 1:
        new_lines = (
            original_lines[: line_number - 1]
            + [content + "\n"]
            + original_lines[line_number - 1 :]
        )
    else:
        return ToolResponse(
            content=[
                TextBlock(
                    type="text",
                    text="InvalidArgumentsError: The given line_number "
                    f"({line_number}) is not in the valid range "
                    f"[1, {len(original_lines) + 1}].",
                ),
            ],
        )

    with open(file_path, "w", encoding="utf-8") as file:
        file.writelines(new_lines)

    with open(file_path, "r", encoding="utf-8") as file:
        new_lines = file.readlines()

    start, end = _calculate_view_ranges(
        len(original_lines),
        len(new_lines),
        line_number,
        line_number,
        extra_view_n_lines=5,
    )

    show_content = _view_text_file(file_path, [start, end])

    return ToolResponse(
        content=[
            TextBlock(
                type="text",
                text=f"Insert content into {file_path} at line "
                f"{line_number} successfully. The new content "
                f"between lines {start}-{end} is:\n"
                f"```\n{show_content}```",
            ),
        ],
    )


async def write_text_file(
    file_path: str,
    content: str,
    ranges: None | list[int] = None,
) -> ToolResponse:
    """Create/Replace/Overwrite content in a text file. When `ranges` is provided, the content will be replaced in the specified range. Otherwise, the entire file (if exists) will be overwritten.

    Args:
        file_path (`str`):
            The target file path.
        content (`str`):
            The content to be written.
        ranges (`list[int] | None`, defaults to `None`):
            The range of lines to be replaced. If `None`, the entire file will
            be overwritten.

    Returns:
        `ToolResponse`:
            The tool response containing the result of the writing operation.
    """

    if not os.path.exists(file_path):
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(content)

        if ranges:
            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Create and write {file_path} successfully. "
                        f"The ranges {ranges} is ignored because the "
                        f"file does not exist.",
                    ),
                ],
            )

        return ToolResponse(
            content=[
                TextBlock(
                    type="text",
                    text=f"Create and write {file_path} successfully.",
                ),
            ],
        )

    with open(file_path, "r", encoding="utf-8") as file:
        original_lines = file.readlines()

    if ranges is not None:
        if (
            isinstance(ranges, list)
            and len(ranges) == 2
            and all(isinstance(i, int) for i in ranges)
        ):
            # Replace content in the specified range
            start, end = ranges
            if start > len(original_lines):
                return ToolResponse(
                    content=[
                        TextBlock(
                            type="text",
                            text=f"Error: The start line {start} is invalid. "
                            f"The file only has {len(original_lines)} "
                            f"lines.",
                        ),
                    ],
                )

            new_content = (
                original_lines[: start - 1]
                + [
                    content,
                ]
                + original_lines[end:]
            )

            with open(file_path, "w", encoding="utf-8") as file:
                file.write("".join(new_content))

            # The written content may contain multiple "\n", to avoid mis
            # counting the lines, we read the file again to get the new content
            with open(file_path, "r", encoding="utf-8") as file:
                new_lines = file.readlines()

            view_start, view_end = _calculate_view_ranges(
                len(original_lines),
                len(new_lines),
                start,
                end,
            )

            content = "".join(
                [
                    f"{index + view_start}: {line}"
                    for index, line in enumerate(
                        new_lines[view_start - 1 : view_end],
                    )
                ],
            )

            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"""Write {file_path} successfully. The new content snippet:
```
{content}```""",
                    ),
                ],
            )

        else:
            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Error: Invalid range format. Expected a list "
                        f"of two integers, but got {ranges}.",
                    ),
                ],
            )

    with open(file_path, "w", encoding="utf-8") as file:
        file.write(content)

    return ToolResponse(
        content=[
            TextBlock(
                type="text",
                text=f"Overwrite {file_path} successfully.",
            ),
        ],
    )



# -*- coding: utf-8 -*-
"""The embedding model base class."""
from typing import Any

from ._embedding_response import EmbeddingResponse


class EmbeddingModelBase:
    """Base class for embedding models."""

    model_name: str
    """The embedding model name"""

    supported_modalities: list[str]
    """The supported data modalities, e.g. "text", "image", "video"."""

    dimensions: int
    """The dimensions of the embedding vector."""

    def __init__(
        self,
        model_name: str,
        dimensions: int,
    ) -> None:
        """Initialize the embedding model base class.

        Args:
            model_name (`str`):
                The name of the embedding model.
            dimensions (`int`):
                The dimension of the embedding vector.
        """
        self.model_name = model_name
        self.dimensions = dimensions

    async def __call__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> EmbeddingResponse:
        """Call the embedding API with the given arguments."""
        raise NotImplementedError(
            f"The {self.__class__.__name__} class does not implement "
            f"the __call__ method.",
        )



# -*- coding: utf-8 -*-
"""The embedding module in agentscope."""

from ._embedding_base import EmbeddingModelBase
from ._embedding_usage import EmbeddingUsage
from ._embedding_response import EmbeddingResponse
from ._dashscope_embedding import DashScopeTextEmbedding
from ._dashscope_multimodal_embedding import DashScopeMultiModalEmbedding
from ._openai_embedding import OpenAITextEmbedding
from ._gemini_embedding import GeminiTextEmbedding
from ._ollama_embedding import OllamaTextEmbedding
from ._cache_base import EmbeddingCacheBase
from ._file_cache import FileEmbeddingCache


__all__ = [
    "EmbeddingModelBase",
    "EmbeddingUsage",
    "EmbeddingResponse",
    "DashScopeTextEmbedding",
    "DashScopeMultiModalEmbedding",
    "OpenAITextEmbedding",
    "GeminiTextEmbedding",
    "OllamaTextEmbedding",
    "EmbeddingCacheBase",
    "FileEmbeddingCache",
]



# -*- coding: utf-8 -*-
"""The runtime configuration in agentscope.

.. note:: You should import this module as ``import ._config``, then use the
 variables defined in this module, instead of ``from ._config import xxx``.
 Because when the variables are changed, the changes will not be reflected in
 the imported module.
"""
from datetime import datetime

import shortuuid


def _generate_random_suffix(length: int) -> str:
    """Generate a random suffix."""
    return shortuuid.uuid()[:length]


project = "UnnamedProject_At" + datetime.now().strftime("%Y%m%d")
name = datetime.now().strftime("%H%M%S_") + _generate_random_suffix(4)
run_id: str = shortuuid.uuid()
created_at: str = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
trace_enabled: bool = False



# -*- coding: utf-8 -*-
"""The truncated formatter base class, which allows to truncate the input
messages."""
from abc import ABC
from copy import deepcopy
from typing import (
    Any,
    Tuple,
    Literal,
    AsyncGenerator,
)

from ._formatter_base import FormatterBase
from ..message import Msg
from ..token import TokenCounterBase
from ..tracing import trace_format


class TruncatedFormatterBase(FormatterBase, ABC):
    """Base class for truncated formatters, which formats input messages into
    required formats with tokens under a specified limit."""

    def __init__(
        self,
        token_counter: TokenCounterBase | None = None,
        max_tokens: int | None = None,
    ) -> None:
        """Initialize the TruncatedFormatterBase.

        Args:
            token_counter (`TokenCounterBase | None`, optional):
                A token counter instance used to count tokens in the messages.
                If not provided, the formatter will format the messages
                without considering token limits.
            max_tokens (`int | None`, optional):
                The maximum number of tokens allowed in the formatted
                messages. If not provided, the formatter will not truncate
                the messages.
        """
        self.token_counter = token_counter

        assert (
            max_tokens is None or 0 < max_tokens
        ), "max_tokens must be greater than 0"
        self.max_tokens = max_tokens

    @trace_format
    async def format(
        self,
        msgs: list[Msg],
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Format the input messages into the required format. If token
        counter and max token limit are provided, the messages will be
        truncated to fit the limit.

        Args:
            msgs (`list[Msg]`):
                The input messages to be formatted.

        Returns:
            `list[dict[str, Any]]`:
                The formatted messages in the required format.
        """

        # Check if the input messages are valid
        self.assert_list_of_msgs(msgs)

        msgs = deepcopy(msgs)

        while True:
            formatted_msgs = await self._format(msgs)
            n_tokens = await self._count(formatted_msgs)

            if (
                n_tokens is None
                or self.max_tokens is None
                or n_tokens <= self.max_tokens
            ):
                return formatted_msgs

            # truncate the input messages
            msgs = await self._truncate(msgs)

    async def _format(self, msgs: list[Msg]) -> list[dict[str, Any]]:
        """Format the input messages into the required format. This method
        should be implemented by the subclasses."""

        formatted_msgs = []
        start_index = 0
        if len(msgs) > 0 and msgs[0].role == "system":
            formatted_msgs.append(
                await self._format_system_message(msgs[0]),
            )
            start_index = 1

        is_first_agent_message = True
        async for typ, group in self._group_messages(msgs[start_index:]):
            match typ:
                case "tool_sequence":
                    formatted_msgs.extend(
                        await self._format_tool_sequence(group),
                    )
                case "agent_message":
                    formatted_msgs.extend(
                        await self._format_agent_message(
                            group,
                            is_first_agent_message,
                        ),
                    )
                    is_first_agent_message = False

        return formatted_msgs

    async def _format_system_message(
        self,
        msg: Msg,
    ) -> dict[str, Any]:
        """Format system message for the LLM API.

        .. note:: This is the default implementation. For certain LLM APIs
        with specific requirements, you may need to implement a custom
        formatting function to accommodate those particular needs.
        """
        return {
            "role": "system",
            "content": msg.get_content_blocks("text"),
        }

    async def _format_tool_sequence(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Given a sequence of tool call/result messages, format them into
        the required format for the LLM API."""
        raise NotImplementedError(
            "_format_tool_sequence is not implemented",
        )

    async def _format_agent_message(
        self,
        msgs: list[Msg],
        is_first: bool = True,
    ) -> list[dict[str, Any]]:
        """Given a sequence of messages without tool calls/results, format
        them into the required format for the LLM API."""
        raise NotImplementedError(
            "_format_agent_message is not implemented",
        )

    async def _truncate(self, msgs: list[Msg]) -> list[Msg]:
        """Truncate the input messages, so that it can fit the token limit.
        This function is called only when

        - both `token_counter` and `max_tokens` are provided,
        - the formatted output of the input messages exceeds the token limit.

        .. tip:: This function only provides a simple strategy, and developers
         can override this method to implement more sophisticated
         truncation strategies.

        .. note:: The tool call message should be truncated together with
         its corresponding tool result message to satisfy the LLM API
         requirements.

        Args:
            msgs (`list[Msg]`):
                The input messages to be truncated.

        Raises:
            `ValueError`:
                If the system prompt message already exceeds the token limit,
                or if there are tool calls without corresponding tool results.

        Returns:
            `list[Msg]`:
                The truncated messages.
        """
        start_index = 0
        if len(msgs) > 0 and msgs[0].role == "system":
            if len(msgs) == 1:
                # If the system prompt already exceeds the token limit, we
                # raise an error.
                raise ValueError(
                    f"The system prompt message already exceeds the token "
                    f"limit ({self.max_tokens} tokens).",
                )

            start_index = 1

        # Create a tool call IDs queues to delete the corresponding tool
        # result message
        tool_call_ids = set()
        for i in range(start_index, len(msgs)):
            msg = msgs[i]
            for block in msg.get_content_blocks("tool_use"):
                tool_call_ids.add(block["id"])

            for block in msg.get_content_blocks("tool_result"):
                try:
                    tool_call_ids.remove(block["id"])
                except KeyError:
                    pass

            # We can stop truncating if the queue is empty
            if len(tool_call_ids) == 0:
                return msgs[:start_index] + msgs[i + 1 :]

        if len(tool_call_ids) > 0:
            raise ValueError(
                "The input messages contains tool call(s) that do not have "
                f"the corresponding tool result(s): {tool_call_ids}. ",
            )

        return msgs[:start_index]

    async def _count(self, msgs: list[dict[str, Any]]) -> int | None:
        """Count the number of tokens in the input messages. If token counter
        is not provided, `None` will be returned.

        Args:
            msgs (`list[Msg]`):
                The input messages to count tokens for.
        """
        if self.token_counter is None:
            return None

        return await self.token_counter.count(msgs)

    @staticmethod
    async def _group_messages(
        msgs: list[Msg],
    ) -> AsyncGenerator[
        Tuple[Literal["tool_sequence", "agent_message"], list[Msg]],
        None,
    ]:
        """Group the input messages into two types and yield them as a
        generator. The two types are:

        - agent message that doesn't contain tool calls/results, and
        - tool sequence that consisted of a sequence of tool calls/results

        .. note:: The group operation is used in multi-agent scenario, where
         multiple entities are involved in the input messages. So that to be
         compatible with tools API, we have to group the messages and format
         them with different strategies.

        Args:
            msgs (`list[Msg]`):
                The input messages to be grouped, where the system prompt
                message shouldn't be included.

        Yields:
            `AsyncGenerator[Tuple[str, list[Msg]], None]`:
                A generator that yields tuples of group type and the list of
                messages in that group. The group type can be either
                "tool_sequence" or "agent_message".
        """

        group_type: Literal["tool_sequence", "agent_message"] | None = None
        group = []
        for msg in msgs:
            if group_type is None:
                if msg.has_content_blocks(
                    "tool_use",
                ) or msg.has_content_blocks("tool_result"):
                    group_type = "tool_sequence"
                else:
                    group_type = "agent_message"

                group.append(msg)
                continue

            # determine if this msg has the same type as the current group
            if group_type == "tool_sequence":
                if msg.has_content_blocks(
                    "tool_use",
                ) or msg.has_content_blocks("tool_result"):
                    group.append(msg)

                else:
                    yield group_type, group
                    group = [msg]
                    group_type = "agent_message"

            elif group_type == "agent_message":
                if msg.has_content_blocks(
                    "tool_use",
                ) or msg.has_content_blocks("tool_result"):
                    yield group_type, group
                    group = [msg]
                    group_type = "tool_sequence"

                else:
                    group.append(msg)
        if group_type:
            yield group_type, group



# -*- coding: utf-8 -*-
"""The user agent class."""
from typing import Type, Any

from pydantic import BaseModel

from ._agent_base import AgentBase
from ._user_input import UserInputBase, TerminalUserInput
from ..message import Msg


class UserAgent(AgentBase):
    """The class for user interaction, allowing developers to handle the user
    input from different sources, such as web UI, cli, and other interfaces.
    """

    _input_method: UserInputBase = TerminalUserInput()
    """The user input method, can be overridden by calling the
    `register_instance/class_input_method` function."""

    def __init__(
        self,
        name: str,
    ) -> None:
        """Initialize the user agent with a name."""
        super().__init__()

        self.name = name

    async def reply(
        self,
        msg: Msg | list[Msg] | None = None,
        structured_model: Type[BaseModel] | None = None,
    ) -> Msg:
        """Receive input message(s) and generate a reply message from the user.

        Args:
            msg (`Msg | list[Msg] | None`, defaults to `None`):
                The message(s) to be replied. If `None`, the agent will wait
                for user input.
            structured_model (`Type[BaseModel] | None`, defaults to `None`):
                A child class of `pydantic.BaseModel` that defines the
                structured output format. If provided, the user will be
                prompted to fill in the required fields.

        Returns:
            `Msg`:
                The reply message generated by the user.
        """

        # Get the input from the specified input method.
        input_data = await self._input_method(
            agent_id=self.id,
            agent_name=self.name,
            structured_model=structured_model,
        )

        blocks_input = input_data.blocks_input
        if (
            blocks_input
            and len(blocks_input) == 1
            and blocks_input[0].get("type") == "text"
        ):
            # Turn blocks_input into a string if only one text block exists
            blocks_input = blocks_input[0].get("text")

        msg = Msg(
            self.name,
            content=blocks_input,
            role="user",
            metadata=input_data.structured_input,
        )

        await self.print(msg)

        return msg

    def override_instance_input_method(
        self,
        input_method: UserInputBase,
    ) -> None:
        """Override the input method of the current UserAgent instance.

        Args:
            input_method (`UserInputBase`):
                The callable input method, which should be an object of a
                class that inherits from `UserInputBase`.
        """
        if not isinstance(input_method, UserInputBase):
            raise ValueError(
                f"The input method should be an instance of the child class "
                f"of `UserInputBase`, but got {type(input_method)} instead.",
            )
        self._input_method = input_method

    @classmethod
    def override_class_input_method(
        cls,
        input_method: UserInputBase,
    ) -> None:
        """Override the input method of the current UserAgent class.

        Args:
            input_method (`UserInputBase`):
                The callable input method, which should be an object of a
                class that inherits from `UserInputBase`.
        """
        if not isinstance(input_method, UserInputBase):
            raise ValueError(
                f"The input method should be an instance of the child class "
                f"of `UserInputBase`, but got {type(input_method)} instead.",
            )
        cls._input_method = input_method

    async def handle_interrupt(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Msg:
        """The post-processing logic when the reply is interrupted by the
        user or something else."""
        raise NotImplementedError(
            f"The handle_interrupt function is not implemented in "
            f"{self.__class__.__name__}",
        )

    async def observe(self, msg: Msg | list[Msg] | None) -> None:
        """Observe the message(s) from the other agents or the environment."""



# -*- coding: utf-8 -*-
"""The tool related types"""

from typing import (
    Callable,
    Union,
    Awaitable,
    AsyncGenerator,
    Generator,
    Coroutine,
    Any,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from ..tool import ToolResponse
else:
    ToolResponse = "ToolResponse"

ToolFunction = Callable[
    ...,
    Union[
        # sync function
        ToolResponse,
        # async function
        Awaitable[ToolResponse],
        # sync generator function
        Generator[ToolResponse, None, None],
        # async generator function
        AsyncGenerator[ToolResponse, None],
        # async function that returns async generator
        Coroutine[Any, Any, AsyncGenerator[ToolResponse, None]],
        # async function that returns sync generator
        Coroutine[Any, Any, Generator[ToolResponse, None, None]],
    ],
]



# -*- coding: utf-8 -*-
"""A file embedding cache implementation for storing and retrieving
embeddings in binary files."""
import hashlib
import json
import os
from typing import Any, List

import numpy as np

from ._cache_base import EmbeddingCacheBase
from .._logging import logger
from ..types import (
    Embedding,
    JSONSerializableObject,
)


class FileEmbeddingCache(EmbeddingCacheBase):
    """The embedding cache class that stores each embeddings vector in
    binary files."""

    def __init__(
        self,
        cache_dir: str = "./.cache/embeddings",
        max_file_number: int | None = None,
        max_cache_size: int | None = None,
    ) -> None:
        """Initialize the file embedding cache class.

        Args:
            cache_dir (`str`, defaults to `"./.cache/embeddings"`):
                The directory to store the embedding files.
            max_file_number (`int | None`, defaults to `None`):
                The maximum number of files to keep in the cache directory. If
                exceeded, the oldest files will be removed.
            max_cache_size (`int | None`, defaults to `None`):
                The maximum size of the cache directory in MB. If exceeded,
                the oldest files will be removed until the size is within the
                limit.
        """
        self._cache_dir = os.path.abspath(cache_dir)
        self.max_file_number = max_file_number
        self.max_cache_size = max_cache_size

    @property
    def cache_dir(self) -> str:
        """The cache directory where the embedding files are stored."""
        if not os.path.exists(self._cache_dir):
            os.makedirs(self._cache_dir, exist_ok=True)
        return self._cache_dir

    async def store(
        self,
        embeddings: List[Embedding],
        identifier: JSONSerializableObject,
        overwrite: bool = False,
        **kwargs: Any,
    ) -> None:
        """Store the embeddings with the given identifier.

        Args:
            embeddings (`List[Embedding]`):
                The embeddings to store.
            identifier (`JSONSerializableObject`):
                The identifier to distinguish the embeddings, which will be
                used to generate a hashable filename, so it should be
                JSON serializable (e.g. a string, number, list, dict).
            overwrite (`bool`, defaults to `False`):
                Whether to overwrite existing embeddings with the same
                identifier. If `True`, existing embeddings will be replaced.
        """
        filename = self._get_filename(identifier)
        path_file = os.path.join(self.cache_dir, filename)

        if os.path.exists(path_file):
            if not os.path.isfile(path_file):
                raise RuntimeError(
                    f"Path {path_file} exists but is not a file.",
                )

            if overwrite:
                np.save(path_file, embeddings)
                await self._maintain_cache_dir()
        else:
            np.save(path_file, embeddings)
            await self._maintain_cache_dir()

    async def retrieve(
        self,
        identifier: JSONSerializableObject,
    ) -> List[Embedding] | None:
        """Retrieve the embeddings with the given identifier. If not found,
        return `None`.

        Args:
            identifier (`JSONSerializableObject`):
                The identifier to retrieve the embeddings, which will be
                used to generate a hashable filename, so it should be
                JSON serializable (e.g. a string, number, list, dict).
        """
        filename = self._get_filename(identifier)
        path_file = os.path.join(self.cache_dir, filename)

        if os.path.exists(path_file):
            return np.load(os.path.join(self.cache_dir, filename)).tolist()
        return None

    async def remove(self, identifier: JSONSerializableObject) -> None:
        """Remove the embeddings with the given identifier.

        Args:
            identifier (`JSONSerializableObject`):
                The identifiers to remove the embeddings, which will be
                used to generate a hashable filename, so it should be
                JSON serializable (e.g. a string, number, list, dict).
        """
        filename = self._get_filename(identifier)
        path_file = os.path.join(self.cache_dir, filename)

        if os.path.exists(path_file):
            os.remove(path_file)
        else:
            raise FileNotFoundError(f"File {path_file} does not exist.")

    async def clear(self) -> None:
        """Clear the cache directory by removing all files."""
        for filename in os.listdir(self.cache_dir):
            if filename.endswith(".npy"):
                os.remove(os.path.join(self.cache_dir, filename))

    def _get_cache_size(self) -> float:
        """Get the current size of the cache directory in MB."""
        total_size = 0
        for filename in os.listdir(self.cache_dir):
            if filename.endswith(".npy"):
                path_file = os.path.join(self.cache_dir, filename)
                if os.path.isfile(path_file):
                    total_size += os.path.getsize(path_file)
        return total_size / (1024.0 * 1024.0)

    @staticmethod
    def _get_filename(identifier: JSONSerializableObject) -> str:
        """Generate a filename based on the identifier."""
        json_str = json.dumps(identifier, ensure_ascii=False)
        return hashlib.sha256(json_str.encode("utf-8")).hexdigest() + ".npy"

    async def _maintain_cache_dir(self) -> None:
        """Maintain the cache directory by removing old files if the number of
        files exceeds the maximum limit or if the cache size exceeds the
        maximum size."""
        files = [
            (_.name, _.stat().st_mtime)
            for _ in os.scandir(self.cache_dir)
            if _.is_file() and _.name.endswith(".npy")
        ]
        files.sort(key=lambda x: x[1])

        if self.max_file_number and len(files) > self.max_file_number:
            for file_name, _ in files[: 0 - self.max_file_number]:
                os.remove(os.path.join(self.cache_dir, file_name))
                logger.info(
                    "Remove cached embedding file %s for limited number "
                    "of files (%d).",
                    file_name,
                    self.max_file_number,
                )
            files = files[0 - self.max_file_number :]

        if (
            self.max_cache_size is not None
            and self._get_cache_size() > self.max_cache_size
        ):
            removed_files = []
            for filename, _ in files:
                os.remove(os.path.join(self.cache_dir, filename))
                removed_files.append(filename)
                if self._get_cache_size() <= self.max_cache_size:
                    break

            if removed_files:
                logger.info(
                    "Remove %d cached embedding file(s) for limited "
                    "cache size (%d MB).",
                    len(removed_files),
                    self.max_cache_size,
                )



# -*- coding: utf-8 -*-
"""The embedding usage class in agentscope."""
from dataclasses import dataclass, field
from typing import Literal

from .._utils._mixin import DictMixin


@dataclass
class EmbeddingUsage(DictMixin):
    """The usage of an embedding model API invocation."""

    time: float
    """The time used in seconds."""

    tokens: int | None = field(default_factory=lambda: None)
    """The number of tokens used, if available."""

    type: Literal["embedding"] = field(default_factory=lambda: "embedding")
    """The type of the usage, must be `embedding`."""



# -*- coding: utf-8 -*-
"""The reader base class for retrieval-augmented generation (RAG)."""
from abc import abstractmethod
from typing import Any

from .._document import Document


class ReaderBase:
    """The reader base class, which is responsible for reading the original
    data, splitting it into chunks, and converting each chunk into a `Document`
    object."""

    @abstractmethod
    async def __call__(self, *args: Any, **kwargs: Any) -> list[Document]:
        """The async call function that takes the input files and returns the
        vector records"""

    @abstractmethod
    def get_doc_id(self, *args: Any, **kwargs: Any) -> str:
        """Get a unique document ID for the input data. This method is to
        expose the document ID generation logic to the developers

        Returns:
            `str`:
                A unique document ID for the input data.
        """



# -*- coding: utf-8 -*-
"""General evaluator implementation in AgentScope, which is easy to debug
compared to the RayEvaluator."""
from typing import Callable, Awaitable, Coroutine, Any

from ._evaluator_base import EvaluatorBase
from .._evaluator_storage import EvaluatorStorageBase
from .._task import Task
from .._solution import SolutionOutput
from .._benchmark_base import BenchmarkBase


class GeneralEvaluator(EvaluatorBase):
    """The general evaluator that support users to debug their evaluation"""

    def __init__(
        self,
        name: str,
        benchmark: BenchmarkBase,
        n_repeat: int,
        storage: EvaluatorStorageBase,
        n_workers: int,
    ) -> None:
        """Initialize the evaluator."""
        super().__init__(
            name=name,
            benchmark=benchmark,
            n_repeat=n_repeat,
            storage=storage,
        )

        assert isinstance(benchmark, BenchmarkBase)

        assert n_repeat >= 1, "n_repeat must be at least 1"

        assert n_workers >= 1, "n_workers must be at least 1"

        self.benchmark = benchmark
        self.n_repeat = n_repeat
        self.n_workers = n_workers

    async def run_evaluation(
        self,
        task: Task,
        repeat_id: str,
        solution_output: SolutionOutput,
    ) -> None:
        """Run the evaluation for a task and solution result."""
        evaluation_results = await task.evaluate(solution_output)
        # store the evaluation result
        for result in evaluation_results:
            self.storage.save_evaluation_result(
                task_id=task.id,
                repeat_id=repeat_id,
                evaluation=result,
            )

    async def run_solution(
        self,
        repeat_id: str,
        task: Task,
        solution: Callable[[Task, Callable], Awaitable[SolutionOutput]],
    ) -> None:
        """Generate a solution to a task and evaluate."""
        if self.storage.solution_result_exists(task.id, repeat_id):
            # Obtain from storage
            solution_result = self.storage.get_solution_result(
                task.id,
                repeat_id,
            )

        else:
            # Run the solution
            solution_result = await solution(
                task,
                self.storage.get_agent_pre_print_hook(
                    task.id,
                    repeat_id,
                ),
            )
            self.storage.save_solution_result(
                task.id,
                repeat_id,
                solution_result,
            )

        # Evaluate the solution with the
        for metric in task.metrics:
            if not self.storage.evaluation_result_exists(
                task.id,
                repeat_id,
                metric.name,
            ):
                await self.run_evaluation(
                    task,
                    repeat_id,
                    solution_result,
                )

    async def run(
        self,
        solution: Callable[
            [Task, Callable],
            Coroutine[Any, Any, SolutionOutput],
        ],
    ) -> None:
        """Run the ray-based distributed and parallel evaluation, and get the
        results.

        Args:
            solution (`Callable[[Task, Callable], Coroutine[Any, Any, \
            SolutionOutput]]`):
                A async function that takes a `Task` instance and a pre-print
                hook function as input, returns a `SolutionOutput` instance.
        """

        await self._save_evaluation_meta()

        for repeat_id in range(self.n_repeat):
            for task in self.benchmark:
                await self.run_solution(
                    str(repeat_id),
                    task,
                    solution,
                )

        await self.aggregate()



# -*- coding: utf-8 -*-
"""The base class for MCP clients in AgentScope."""
from abc import abstractmethod
from typing import Callable, List

import mcp.types

from .._logging import logger
from ..message import ImageBlock, Base64Source, AudioBlock, TextBlock


class MCPClientBase:
    """Base class for MCP clients."""

    def __init__(self, name: str) -> None:
        """Initialize the MCP client with a name.

        Args:
            name (`str`):
                The name to identify the MCP server, which should be unique
                across the MCP servers.
        """
        self.name = name

    @abstractmethod
    async def get_callable_function(
        self,
        func_name: str,
        wrap_tool_result: bool = True,
    ) -> Callable:
        """Get a tool function by its name."""

    @staticmethod
    def _convert_mcp_content_to_as_blocks(
        mcp_content_blocks: list,
    ) -> List[TextBlock | ImageBlock | AudioBlock]:
        """Convert MCP content to AgentScope blocks."""

        as_content: list = []
        for content in mcp_content_blocks:
            if isinstance(content, mcp.types.TextContent):
                as_content.append(
                    TextBlock(
                        type="text",
                        text=content.text,
                    ),
                )
            elif isinstance(content, mcp.types.ImageContent):
                as_content.append(
                    ImageBlock(
                        type="image",
                        source=Base64Source(
                            type="base64",
                            media_type=content.mimeType,
                            data=content.data,
                        ),
                    ),
                )
            elif isinstance(content, mcp.types.AudioContent):
                as_content.append(
                    AudioBlock(
                        type="audio",
                        source=Base64Source(
                            type="base64",
                            media_type=content.mimeType,
                            data=content.data,
                        ),
                    ),
                )
            elif isinstance(content, mcp.types.EmbeddedResource):
                if isinstance(
                    content.resource,
                    mcp.types.TextResourceContents,
                ):
                    as_content.append(
                        TextBlock(
                            type="text",
                            text=content.resource.model_dump_json(indent=2),
                        ),
                    )
                else:
                    # TODO: support the BlobResourceContents in the future,
                    #  which is a base64-encoded string representing the
                    #  binary data
                    logger.error(
                        "Unsupported EmbeddedResource content type: %s. "
                        "Skipping this content.",
                        type(content.resource),
                    )
            else:
                logger.warning(
                    "Unsupported content type: %s. Skipping this content.",
                    type(content),
                )
        return as_content



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""The shell command tool in agentscope."""

import asyncio
from typing import Any

from .._response import ToolResponse
from ...message import TextBlock


async def execute_shell_command(
    command: str,
    timeout: int = 300,
    **kwargs: Any,
) -> ToolResponse:
    """Execute given command and return the return code, standard output and
    error within <returncode></returncode>, <stdout></stdout> and
    <stderr></stderr> tags.

    Args:
        command (`str`):
            The shell command to execute.
        timeout (`float`, defaults to `300`):
            The maximum time (in seconds) allowed for the command to run.

    Returns:
        `ToolResponse`:
            The tool response containing the return code, standard output, and
            standard error of the executed command.
    """

    proc = await asyncio.create_subprocess_shell(
        command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        bufsize=0,
    )

    try:
        await asyncio.wait_for(proc.wait(), timeout=timeout)
        stdout, stderr = await proc.communicate()
        stdout_str = stdout.decode("utf-8")
        stderr_str = stderr.decode("utf-8")
        returncode = proc.returncode

    except asyncio.TimeoutError:
        stderr_suffix = (
            f"TimeoutError: The command execution exceeded "
            f"the timeout of {timeout} seconds."
        )
        returncode = -1
        try:
            proc.terminate()
            stdout, stderr = await proc.communicate()
            stdout_str = stdout.decode("utf-8")
            stderr_str = stderr.decode("utf-8")
            if stderr_str:
                stderr_str += f"\n{stderr_suffix}"
            else:
                stderr_str = stderr_suffix
        except ProcessLookupError:
            stdout_str = ""
            stderr_str = stderr_suffix

    return ToolResponse(
        content=[
            TextBlock(
                type="text",
                text=(
                    f"<returncode>{returncode}</returncode>"
                    f"<stdout>{stdout_str}</stdout>"
                    f"<stderr>{stderr_str}</stderr>"
                ),
            ),
        ],
    )



# -*- coding: utf-8 -*-
# pylint: disable=too-many-branches
"""The Ollama formatter module."""
import base64
import os
from typing import Any
from urllib.parse import urlparse

from ._truncated_formatter_base import TruncatedFormatterBase
from .._logging import logger
from .._utils._common import _get_bytes_from_web_url
from ..message import Msg, TextBlock, ImageBlock, ToolUseBlock, ToolResultBlock
from ..token import TokenCounterBase


def _convert_ollama_image_url_to_base64_data(url: str) -> str:
    """Convert image url to base64."""
    parsed_url = urlparse(url)

    if not os.path.exists(url) and parsed_url.scheme != "":
        # Web url
        data = _get_bytes_from_web_url(url)
        return data
    if os.path.exists(url):
        # Local file
        with open(url, "rb") as f:
            data = base64.b64encode(f.read()).decode("utf-8")

        return data

    raise ValueError(
        f"The URL `{url}` is not a valid image URL or local file.",
    )


class OllamaChatFormatter(TruncatedFormatterBase):
    """Formatter for Ollama messages."""

    support_tools_api: bool = True
    """Whether support tools API"""

    support_multiagent: bool = False
    """Whether support multi-agent conversations"""

    support_vision: bool = True
    """Whether support vision data"""

    supported_blocks: list[type] = [
        TextBlock,
        # Multimodal
        ImageBlock,
        # Tool use
        ToolUseBlock,
        ToolResultBlock,
    ]
    """The list of supported message blocks"""

    async def _format(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Format message objects into Ollama API format.

        Args:
            msgs (`list[Msg]`):
                The list of message objects to format.

        Returns:
            `list[dict[str, Any]]`:
                The formatted messages as a list of dictionaries.
        """
        self.assert_list_of_msgs(msgs)

        messages: list[dict] = []
        for msg in msgs:
            content_blocks: list = []
            tool_calls = []
            images = []

            for block in msg.get_content_blocks():
                typ = block.get("type")
                if typ == "text":
                    content_blocks.append({**block})

                elif typ == "tool_use":
                    tool_calls.append(
                        {
                            "id": block.get("id"),
                            "type": "function",
                            "function": {
                                "name": block.get("name"),
                                "arguments": block.get("input", {}),
                            },
                        },
                    )

                elif typ == "tool_result":
                    messages.append(
                        {
                            "role": "tool",
                            "tool_call_id": block.get("id"),
                            "content": self.convert_tool_result_to_string(
                                block.get("output"),  # type: ignore[arg-type]
                            ),
                            "name": block.get("name"),
                        },
                    )

                elif typ == "image":
                    source_type = block["source"]["type"]
                    if source_type == "url":
                        images.append(
                            _convert_ollama_image_url_to_base64_data(
                                block["source"]["url"],
                            ),
                        )
                    elif source_type == "base64":
                        images.append(block["source"]["data"])

                else:
                    logger.warning(
                        "Unsupported block type %s in the message, skipped.",
                        typ,
                    )
            content_msg = "\n".join(
                content.get("text", "") for content in content_blocks
            )
            msg_ollama = {
                "role": msg.role,
                "content": content_msg or None,
            }

            if tool_calls:
                msg_ollama["tool_calls"] = tool_calls

            if images:
                msg_ollama["images"] = images

            if msg_ollama["content"] or msg_ollama.get("tool_calls"):
                messages.append(msg_ollama)

        return messages


class OllamaMultiAgentFormatter(TruncatedFormatterBase):
    """
    Ollama formatter for multi-agent conversations, where more than
    a user and an agent are involved.
    """

    support_tools_api: bool = True
    """Whether support tools API"""

    support_multiagent: bool = True
    """Whether support multi-agent conversations"""

    support_vision: bool = True
    """Whether support vision data"""

    supported_blocks: list[type] = [
        TextBlock,
        # Multimodal
        ImageBlock,
        # Tool use
        ToolUseBlock,
        ToolResultBlock,
    ]
    """The list of supported message blocks"""

    def __init__(
        self,
        conversation_history_prompt: str = (
            "# Conversation History\n"
            "The content between <history></history> tags contains "
            "your conversation history\n"
        ),
        token_counter: TokenCounterBase | None = None,
        max_tokens: int | None = None,
    ) -> None:
        """Initialize the Ollama multi-agent formatter.

        Args:
            conversation_history_prompt (`str`):
                The prompt to use for the conversation history section.
            token_counter (`TokenCounterBase | None`, optional):
                The token counter used for truncation.
            max_tokens (`int | None`, optional):
                The maximum number of tokens allowed in the formatted
                messages. If `None`, no truncation will be applied.
        """
        super().__init__(token_counter=token_counter, max_tokens=max_tokens)
        self.conversation_history_prompt = conversation_history_prompt

    async def _format_system_message(
        self,
        msg: Msg,
    ) -> dict[str, Any]:
        """Format system message for the Ollama API."""
        return {
            "role": "system",
            "content": msg.get_text_content(),
        }

    async def _format_tool_sequence(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Given a sequence of tool call/result messages, format them into
        the required format for the Ollama API.

        Args:
            msgs (`list[Msg]`):
                The list of messages containing tool calls/results to format.

        Returns:
            `list[dict[str, Any]]`:
                A list of dictionaries formatted for the Ollama API.
        """
        return await OllamaChatFormatter().format(msgs)

    async def _format_agent_message(
        self,
        msgs: list[Msg],
        is_first: bool = True,
    ) -> list[dict[str, Any]]:
        """Given a sequence of messages without tool calls/results, format
        them into the required format for the Ollama API.

        Args:
            msgs (`list[Msg]`):
                A list of Msg objects to be formatted.
            is_first (`bool`, defaults to `True`):
                Whether this is the first agent message in the conversation.
                If `True`, the conversation history prompt will be included.

        Returns:
            `list[dict[str, Any]]`:
                A list of dictionaries formatted for the ollama API.
        """

        if is_first:
            conversation_history_prompt = self.conversation_history_prompt
        else:
            conversation_history_prompt = ""

        # Format into required Ollama format
        formatted_msgs: list[dict] = []

        # Collect the multimodal files
        conversation_blocks: list = []
        accumulated_text = []
        images = []
        for msg in msgs:
            for block in msg.get_content_blocks():
                if block["type"] == "text":
                    accumulated_text.append(f"{msg.name}: {block['text']}")

                elif block["type"] == "image":
                    # Handle the accumulated text as a single block
                    source = block["source"]
                    if accumulated_text:
                        conversation_blocks.append(
                            {"text": "\n".join(accumulated_text)},
                        )
                        accumulated_text.clear()

                    if source["type"] == "url":
                        images.append(
                            _convert_ollama_image_url_to_base64_data(
                                source["url"],
                            ),
                        )

                    elif source["type"] == "base64":
                        images.append(source["data"])

                    conversation_blocks.append({**block})

        if accumulated_text:
            conversation_blocks.append(
                {"text": "\n".join(accumulated_text)},
            )

        if conversation_blocks:
            if conversation_blocks[0].get("text"):
                conversation_blocks[0]["text"] = (
                    conversation_history_prompt
                    + "<history>\n"
                    + conversation_blocks[0]["text"]
                )

            else:
                conversation_blocks.insert(
                    0,
                    {
                        "text": conversation_history_prompt + "<history>\n",
                    },
                )

            if conversation_blocks[-1].get("text"):
                conversation_blocks[-1]["text"] += "\n</history>"

            else:
                conversation_blocks.append({"text": "</history>"})

        conversation_blocks_text = "\n".join(
            conversation_block.get("text", "")
            for conversation_block in conversation_blocks
        )

        user_message = {
            "role": "user",
            "content": conversation_blocks_text,
        }
        if images:
            user_message["images"] = images
        if conversation_blocks:
            formatted_msgs.append(user_message)

        return formatted_msgs



# -*- coding: utf-8 -*-
"""
Wrap OpenAI API calls as tools. Refer the official
`OpenAI API documentation <https://platform.openai.com/docs/overview>`_ for
more details.
"""
import base64
from io import BytesIO
import os
from typing import Literal, IO
import requests

from .. import ToolResponse
from ...formatter._openai_formatter import _to_openai_image_url
from ...message import (
    ImageBlock,
    TextBlock,
    Base64Source,
    URLSource,
    AudioBlock,
)


def _parse_url(url: str) -> BytesIO | IO[bytes]:
    """
    If url is a local file path, return a BytesIO of the file content.
    If url is a web URL, fetch the content and return as BytesIO.
    """
    if url.startswith(("http://", "https://")):
        response = requests.get(url)
        response.raise_for_status()  # Raise an exception for HTTP errors
        return BytesIO(response.content)
    else:
        if not os.path.exists(url):
            raise FileNotFoundError(f"File not found: {url}")
        return open(os.path.abspath(url), "rb")


def openai_text_to_image(
    prompt: str,
    api_key: str,
    n: int = 1,
    model: Literal["dall-e-2", "dall-e-3", "gpt-image-1"] = "dall-e-2",
    size: Literal[
        "256x256",
        "512x512",
        "1024x1024",
        "1792x1024",
        "1024x1792",
    ] = "256x256",
    quality: Literal[
        "auto",
        "standard",
        "hd",
        "high",
        "medium",
        "low",
    ] = "auto",
    style: Literal["vivid", "natural"] = "vivid",
    response_format: Literal["url", "b64_json"] = "url",
) -> ToolResponse:
    """
    Generate image(s) based on the given prompt, and return image URL(s) or
    base64 data.

    Args:
        prompt (`str`):
            The text prompt to generate images.
        api_key (`str`):
            The API key for the OpenAI API.
        n (`int`, defaults to `1`):
            The number of images to generate.
        model (`Literal["dall-e-2", "dall-e-3"]`, defaults to `"dall-e-2"`):
            The model to use for image generation.
        size (`Literal["256x256", "512x512", "1024x1024", "1792x1024", \
        "1024x1792"]`, defaults to `"256x256"`):
            The size of the generated images.
            Must be one of 1024x1024, 1536x1024 (landscape), 1024x1536 (
            portrait), or auto (default value) for gpt-image-1,
            one of 256x256, 512x512, or 1024x1024 for dall-e-2,
            and one of 1024x1024, 1792x1024, or 1024x1792 for dall-e-3.
        quality (`Literal["auto", "standard", "hd", "high", "medium", \
        "low"]`,  defaults to `"auto"`):
            The quality of the image that will be generated.

            - `auto` (default value) will automatically select the best
              quality for the given model.
            - `high`, `medium` and `low` are supported for gpt-image-1.
            - `hd` and `standard` are supported for dall-e-3.
            - `standard` is the only option for dall-e-2.
        style (`Literal["vivid", "natural"]`, defaults to `"vivid"`):
            The style of the generated images.
            This parameter is only supported for dall-e-3.
            Must be one of `vivid` or `natural`.

            - `Vivid` causes the model to lean towards generating hyper-real
              and dramatic images.
            - `Natural` causes the model to produce more natural,
              less hyper-real looking images.
        response_format (`Literal["url", "b64_json"]`, defaults to `"url"`):
            The format in which generated images with dall-e-2 and dall-e-3
            are returned.

            - Must be one of "url" or "b64_json".
            - URLs are only valid for 60 minutes after the image has been
              generated.
            - This parameter isn't supported for gpt-image-1 which will always
              return base64-encoded images.
    Returns:
        `ToolResponse`:
            A ToolResponse containing the generated content
            (ImageBlock/TextBlock/AudioBlock) or error information if the
            operation failed.
    """

    kwargs = {
        "model": model,
        "prompt": prompt,
        "n": n,
        "size": size,
    }
    if model == "dall-e-3":
        kwargs["style"] = style
    if model != "dall-e-2":
        kwargs["quality"] = quality
    if model != "gpt-image-1":
        kwargs["response_format"] = response_format
    if model == "gpt-image-1":
        response_format = "b64_json"

    try:
        import openai

        client = openai.OpenAI(
            api_key=api_key,
        )
        response = client.images.generate(
            **kwargs,
        )
        image_blocks: list = []
        if response_format == "url":
            image_urls = [_.url for _ in response.data]
            for image_url in image_urls:
                image_blocks.append(
                    ImageBlock(
                        type="image",
                        source=URLSource(
                            type="url",
                            url=image_url,
                        ),
                    ),
                )
        else:
            image_datas = [_.b64_json for _ in response.data]
            for image_data in image_datas:
                image_blocks.append(
                    ImageBlock(
                        type="image",
                        source=Base64Source(
                            type="base64",
                            media_type="image/png",
                            data=image_data,
                        ),
                    ),
                )
        return ToolResponse(
            content=image_blocks,
        )
    except Exception as e:
        return ToolResponse(
            [
                TextBlock(
                    type="text",
                    text=f"Failed to generate image: {str(e)}",
                ),
            ],
        )


def openai_edit_image(
    image_url: str,
    prompt: str,
    api_key: str,
    model: Literal["dall-e-2", "gpt-image-1"] = "dall-e-2",
    mask_url: str | None = None,
    n: int = 1,
    size: Literal[
        "256x256",
        "512x512",
        "1024x1024",
    ] = "256x256",
    response_format: Literal["url", "b64_json"] = "url",
) -> ToolResponse:
    """
    Edit an image based on the provided mask and prompt, and return the edited
    image URL(s) or base64 data.

    Args:
        image_url (`str`):
            The file path or URL to the image that needs editing.
        prompt (`str`):
            The text prompt describing the edits to be made to the image.
        api_key (`str`):
            The API key for the OpenAI API.
        model (`Literal["dall-e-2", "gpt-image-1"]`, defaults to `"dall-e-2"`):
            The model to use for image generation.
        mask_url (`str | None`, defaults to `None`):
            The file path or URL to the mask image that specifies the regions
            to be edited.
        n (`int`, defaults to `1`):
            The number of edited images to generate.
        size (`Literal["256x256", "512x512", "1024x1024"]`, defaults to \
        `"256x256"`):
            The size of the edited images.
        response_format (`Literal["url", "b64_json"]`, defaults to `"url"`):
            The format in which generated images are returned.

            - Must be one of "url" or "b64_json".
            - URLs are only valid for 60 minutes after generation.
            - This parameter isn't supported for gpt-image-1 which will
              always return base64-encoded images.

    Returns:
        `ToolResponse`:
            A ToolResponse containing the generated content
            (ImageBlock/TextBlock/AudioBlock) or error information if the
            operation failed.

    """

    try:
        import openai

        client = openai.OpenAI(
            api_key=api_key,
        )

        def prepare_image(url_or_path: str) -> BytesIO:
            from PIL import Image

            if url_or_path.startswith(("http://", "https://")):
                response = requests.get(url_or_path)
                response.raise_for_status()
                img = Image.open(BytesIO(response.content))
            else:
                img = Image.open(url_or_path)

            if img.mode != "RGBA":
                img = img.convert("RGBA")
            img_buffer = BytesIO()
            img.save(img_buffer, format="PNG")
            img_buffer.seek(0)
            img_buffer.name = "image.png"
            return img_buffer

        image_file = prepare_image(image_url)

        kwargs = {
            "model": model,
            "image": image_file,
            "prompt": prompt,
            "n": n,
            "size": size,
        }

        if mask_url:
            kwargs["mask"] = prepare_image(mask_url)

        if model == "dall-e-2":
            kwargs["response_format"] = response_format
        else:
            response_format = "b64_json"

        response = client.images.edit(**kwargs)

        if response_format == "url":
            urls = [_.url for _ in response.data]
            image_blocks: list = []
            for url in urls:
                image_blocks.append(
                    ImageBlock(
                        type="image",
                        source=URLSource(
                            type="url",
                            url=url,
                        ),
                    ),
                )
            return ToolResponse(
                content=image_blocks,
            )
        elif response_format == "b64_json":
            image_datas = [_.b64_json for _ in response.data]
            image_blocks = []
            for image_data in image_datas:
                image_blocks.append(
                    ImageBlock(
                        type="image",
                        source=Base64Source(
                            type="base64",
                            media_type="image/png",
                            data=image_data,
                        ),
                    ),
                )
            return ToolResponse(
                content=image_blocks,
            )
    except Exception as e:
        return ToolResponse(
            [
                TextBlock(
                    type="text",
                    text=f"Failed to generate image: {str(e)}",
                ),
            ],
        )


def openai_create_image_variation(
    image_url: str,
    api_key: str,
    n: int = 1,
    model: Literal["dall-e-2"] = "dall-e-2",
    size: Literal[
        "256x256",
        "512x512",
        "1024x1024",
    ] = "256x256",
    response_format: Literal["url", "b64_json"] = "url",
) -> ToolResponse:
    """
    Create variations of an image and return the image URL(s) or base64 data.

    Args:
        image_url (`str`):
            The file path or URL to the image from which variations will be
            generated.
        api_key (`str`):
            The API key for the OpenAI API.
        n (`int`, defaults to `1`):
            The number of image variations to generate.
        model (` Literal["dall-e-2"]`, default to `dall-e-2`):
            The model to use for image variation.
        size (`Literal["256x256", "512x512", "1024x1024"]`, defaults to \
        `"256x256"`):
            The size of the generated image variations.
        response_format (`Literal["url", "b64_json"]`, defaults to `"url"`):
            The format in which generated images are returned.

            - Must be one of url or b64_json.
            - URLs are only valid for 60 minutes after the image has been
              generated.

    Returns:
        `ToolResponse`:
            A ToolResponse containing the generated content
            (ImageBlock/TextBlock/AudioBlock) or error information if the
            operation failed.
    """

    # _parse_url handles both local and web URLs and returns BytesIO
    image = _parse_url(image_url)
    try:
        import openai

        client = openai.OpenAI(
            api_key=api_key,
        )
        response = client.images.create_variation(
            model=model,
            image=image,
            n=n,
            size=size,
        )
        image_blocks: list = []
        if response_format == "url":
            urls = [_.url for _ in response.data]
            for url in urls:
                image_blocks.append(
                    ImageBlock(
                        type="image",
                        source=URLSource(
                            type="url",
                            url=url,
                        ),
                    ),
                )
        else:
            image_datas = [_.b64_json for _ in response.data]
            for image_data in image_datas:
                image_blocks.append(
                    ImageBlock(
                        type="image",
                        source=Base64Source(
                            type="base64",
                            media_type="image/png",
                            data=image_data,
                        ),
                    ),
                )
        return ToolResponse(
            content=image_blocks,
        )
    except Exception as e:
        return ToolResponse(
            [
                TextBlock(
                    type="text",
                    text=f"Failed to generate image: {str(e)}",
                ),
            ],
        )


def openai_image_to_text(
    image_urls: str | list[str],
    api_key: str,
    prompt: str = "Describe the image",
    model: str = "gpt-4o",
) -> ToolResponse:
    """
    Generate descriptive text for given image(s) using a specified model, and
    return the generated text.

    Args:
        image_urls (`str | list[str]`):
            The URL or list of URLs pointing to the images that need to be
            described.
        api_key (`str`):
            The API key for the OpenAI API.
        prompt (`str`, defaults to `"Describe the image"`):
            The prompt that instructs the model on how to describe
            the image(s).
        model (`str`, defaults to `"gpt-4o"`):
            The model to use for generating the text descriptions.

    Returns:
        `ToolResponse`:
            A ToolResponse containing the generated content
            (ImageBlock/TextBlock/AudioBlock) or error information if the
            operation failed.

    """
    if isinstance(image_urls, str):
        image_urls = [image_urls]

    content = []
    for url in image_urls:
        content.append(
            {
                "type": "image_url",
                "image_url": {
                    "url": _to_openai_image_url(url),
                },
            },
        )
    content.append(
        {
            "type": "text",
            "text": prompt,
        },
    )
    messages = [
        {
            "role": "user",
            "content": content,
        },
    ]

    try:
        import openai

        client = openai.OpenAI(
            api_key=api_key,
        )
        response = client.chat.completions.create(
            messages=messages,
            model=model,
        )
        return ToolResponse(
            [
                TextBlock(
                    type="text",
                    text=response.choices[0].message.content,
                ),
            ],
        )
    except Exception as e:
        return ToolResponse(
            [
                TextBlock(
                    type="text",
                    text=f"Failed to generate text: {str(e)}",
                ),
            ],
        )


def openai_text_to_audio(
    text: str,
    api_key: str,
    model: Literal["tts-1", "tts-1-hd", "gpt-4o-mini-tts"] = "tts-1",
    voice: Literal[
        "alloy",
        "ash",
        "ballad",
        "coral",
        "echo",
        "fable",
        "nova",
        "onyx",
        "sage",
        "shimmer",
    ] = "alloy",
    speed: float = 1.0,
    res_format: Literal[
        "mp3",
        "opus",
        "aac",
        "flac",
        "wav",
        "pcm",
    ] = "mp3",
) -> ToolResponse:
    """
    Convert text to an audio file using a specified model and voice.

    Args:
        text (`str`):
            The text to convert to audio.
        api_key (`str`):
            The API key for the OpenAI API.
        model (`Literal["tts-1", "tts-1-hd"]`, defaults to `"tts-1"`):
            The model to use for text-to-speech conversion.
        voice (`Literal["alloy", "echo", "fable", "onyx", "nova", \
        "shimmer"]`, defaults to `"alloy"`):
            The voice to use for the audio output.
        speed (`float`, defaults to `1.0`):
            The speed of the audio playback. A value of 1.0 is normal speed.
        res_format (`Literal["mp3", "wav", "opus", "aac", "flac", \
        "wav", "pcm"]`, defaults to `"mp3"`):
            The format of the audio file.

    Returns:
        `ToolResponse`:
            A ToolResponse containing the generated content
            (ImageBlock/TextBlock/AudioBlock) or error information if the
            operation failed.
    """

    try:
        import openai

        client = openai.OpenAI(
            api_key=api_key,
        )
        response = client.audio.speech.create(
            model=model,
            voice=voice,
            speed=speed,
            input=text,
            response_format=res_format,
        )
        audio_bytes = response.content
        audio_base64 = base64.b64encode(audio_bytes).decode("utf-8")

        return ToolResponse(
            [
                AudioBlock(
                    type="audio",
                    source=Base64Source(
                        type="base64",
                        media_type=f"audio/{res_format}",
                        data=audio_base64,
                    ),
                ),
            ],
        )
    except Exception as e:
        return ToolResponse(
            [
                TextBlock(
                    type="text",
                    text=f"Error: Failed to generate audio. {str(e)}",
                ),
            ],
        )


def openai_audio_to_text(
    audio_file_url: str,
    api_key: str,
    language: str = "en",
    temperature: float = 0.2,
) -> ToolResponse:
    """
    Convert an audio file to text using OpenAI's transcription service.

    Args:
        audio_file_url (`str`):
            The file path or URL to the audio file that needs to be
            transcribed.
        api_key (`str`):
            The API key for the OpenAI API.
        language (`str`, defaults to `"en"`):
            The language of the input audio in
            `ISO-639-1 format \
            <https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes>`_
            (e.g., "en", "zh", "fr"). Improves accuracy and latency.
        temperature (`float`, defaults to `0.2`):
            The temperature for the transcription, which affects the
            randomness of the output.

    Returns:
        `ToolResponse`:
            A ToolResponse containing the generated content
            (ImageBlock/TextBlock/AudioBlock) or error information if the
            operation failed.
    """

    try:
        import openai

        client = openai.OpenAI(
            api_key=api_key,
        )

        if audio_file_url.startswith(("http://", "https://")):
            response = requests.get(audio_file_url)
            response.raise_for_status()
            audio_buffer = BytesIO(response.content)
            import urllib.parse
            from pathlib import Path

            parsed_url = urllib.parse.urlparse(audio_file_url)
            filename = Path(parsed_url.path).name or "audio.mp3"
            audio_buffer.name = filename
            audio_file = audio_buffer
            transcription = client.audio.transcriptions.create(
                model="whisper-1",
                file=audio_file,
                language=language,
                temperature=temperature,
            )
        else:
            if not os.path.exists(audio_file_url):
                raise FileNotFoundError(f"File not found: {audio_file_url}")
            with open(audio_file_url, "rb") as audio_file:
                transcription = client.audio.transcriptions.create(
                    model="whisper-1",
                    file=audio_file,
                    language=language,
                    temperature=temperature,
                )
        return ToolResponse(
            [
                TextBlock(
                    type="text",
                    text=transcription.text,
                ),
            ],
        )
    except Exception as e:
        return ToolResponse(
            [
                TextBlock(
                    type="text",
                    text=f"Error: Failed to transcribe audio: {str(e)}",
                ),
            ],
        )



# -*- coding: utf-8 -*-
"""The reminder API in ACEBench simulation tools."""
from datetime import datetime

from ._shared_state import SharedState


class ReminderApi(SharedState):
    """The reminder Api in the ACEBench evaluation."""

    tool_functions: list[str] = [
        "view_reminder_by_title",
        "add_reminder",
        "delete_reminder",
        "view_all_reminders",
        "mark_as_notified",
        "search_reminders",
    ]

    def __init__(self, share_state: dict) -> None:
        """Initialize the Reminder Api in the ACEBench evaluation."""
        super().__init__(share_state)

        self.max_capacity = 6
        self.reminder_list: dict[
            int,
            dict,
        ] = {
            1: {
                "reminder_id": 1001,
                "title": "Doctor's Appointment",
                "description": "Visit Dr. Smith for a checkup.",
                "time": "2024-07-15 09:30",
                "notified": False,
            },
            2: {
                "reminder_id": 1002,
                "title": "Team Meeting",
                "description": "Monthly project review with the team.",
                "time": "2024-07-17 11:00",
                "notified": False,
            },
            3: {
                "reminder_id": 1003,
                "title": "To-do list",
                "description": '首先帮Frank在"盒马生鲜"点外卖，'
                '需要定两个"生鲜大礼包"，再发短信告诉Frank：'
                '"购买商品的价格是()元"。要把括号换成实际金额，'
                "保留一位小数。",
                "time": "2024-07-16 11:00",
                "notified": False,
            },
        }
        self.reminder_id_counter: int = 3

    def get_state_dict(self) -> dict:
        """Get the current state dict of the ReminderApi."""
        return {
            "ReminderApi": {
                "reminder_list": self.reminder_list,
            },
        }

    def _check_capacity(self) -> bool:
        """检查备忘录容量是否已满。"""
        return len(self.reminder_list) >= self.max_capacity

    def view_reminder_by_title(
        self,
        title: str,
    ) -> dict[str, str | bool | dict[str, str | bool | datetime]]:
        """根据提醒的标题查看特定的提醒。

        Args:
            title (str): 提醒的标题。

        Returns:
            dict[str, str | bool | dict[str, str | bool | datetime]]:
                包含查找状态和提醒详情的字典。
        """
        if not self.logged_in:
            return {"status": False, "message": "device未登录，无法查看提醒"}
        for reminder in self.reminder_list.values():
            if reminder["title"] == title:
                return {"status": True, "reminder": reminder}

        return {"status": False, "message": f"没有找到标题为 '{title}' 的提醒"}

    def add_reminder(
        self,
        title: str,
        description: str,
        time: datetime,
    ) -> dict[str, bool | str]:
        """添加一个新的提醒。

        Args:
            title (str): 提醒标题。
            description (str): 提醒描述。
            time (datetime): 提醒时间, 一定遵循格式"YYYY-MM-DD HH:MM"。

        Returns:
            dict[str, bool | str]: 包含添加状态和结果的字典。
        """
        if not self.logged_in:
            return {
                "status": False,
                "message": "device未登录，无法添加一个新的提醒",
            }
        if self._check_capacity():
            return {"status": False, "message": "提醒容量已满，无法添加新的提醒"}

        self.reminder_id_counter += 1
        reminder_id = self.reminder_id_counter
        self.reminder_list[reminder_id] = {
            "reminder_id": reminder_id,
            "title": title,
            "description": description,
            "time": time,
            "notified": False,
        }
        return {"status": True, "message": f"提醒 '{title}' 已成功添加"}

    def delete_reminder(self, reminder_id: int) -> dict[str, bool | str]:
        """删除指定的提醒。

        Args:
            reminder_id (int): 要删除的提醒ID。

        Returns:
            dict[str, bool | str]: 包含删除状态和结果的字典。
        """
        if not self.logged_in:
            return {"status": False, "message": "device未登录，无法删除指定的提醒"}
        if reminder_id not in self.reminder_list:
            return {"status": False, "message": "提醒ID不存在"}

        del self.reminder_list[reminder_id]
        return {"status": True, "message": f"提醒ID {reminder_id} 已成功删除"}

    def view_all_reminders(
        self,
    ) -> dict:
        """查看所有的提醒。

        Returns:
            dict:
                包含所有提醒的字典列表。
        """
        if not self.reminder_list:
            return {"status": False, "message": "没有任何提醒"}

        reminders = []
        for reminder in self.reminder_list.values():
            reminders.append(
                {
                    "title": reminder["title"],
                    "description": reminder["description"],
                    "time": reminder["time"],
                    "notified": reminder["notified"],
                },
            )
        return {"status": True, "reminders": reminders}

    def mark_as_notified(
        self,
        reminder_id: int,
    ) -> dict[str, bool | str]:
        """标记提醒为已通知。

        Args:
            reminder_id (int): 要标记为已通知的提醒ID。

        Returns:
            dict[str, bool | str]:: 包含操作结果的字典。
        """
        if reminder_id not in self.reminder_list:
            return {"status": False, "message": "提醒ID不存在"}

        self.reminder_list[reminder_id]["notified"] = True
        return {"status": True, "message": f"提醒ID {reminder_id} 已标记为已通知"}

    def search_reminders(
        self,
        keyword: str,
    ) -> dict:
        """根据关键词搜索提醒。

        Args:
            keyword (str): 搜索关键词。

        Returns:
            `dict`:
                包含匹配提醒的字典列表。
        """
        matched_reminders = []

        for reminder in self.reminder_list.values():
            if (
                keyword.lower() in reminder["title"].lower()
                or keyword.lower() in reminder["description"].lower()
            ):
                matched_reminders.append(
                    {
                        "title": reminder["title"],
                        "description": reminder["description"],
                        "time": reminder["time"].strftime("%Y-%m-%d %H:%M"),
                    },
                )

        if not matched_reminders:
            return {"status": False, "message": "没有找到包含该关键词的提醒"}

        return {"status": True, "reminders": matched_reminders}



# -*- coding: utf-8 -*-
"""The dashscope multimodal embedding model in agentscope."""
from datetime import datetime
from typing import Any, Literal

from ._cache_base import EmbeddingCacheBase
from ._embedding_response import EmbeddingResponse
from ._embedding_usage import EmbeddingUsage
from ._embedding_base import EmbeddingModelBase
from ..message import (
    VideoBlock,
    ImageBlock,
    TextBlock,
)


class DashScopeMultiModalEmbedding(EmbeddingModelBase):
    """The DashScope multimodal embedding API, supporting text, image and
    video embedding."""

    supported_modalities: list[str] = ["text", "image", "video"]
    """This class supports text, image and video input."""

    def __init__(
        self,
        api_key: str,
        model_name: str,
        dimensions: int | None = None,
        embedding_cache: EmbeddingCacheBase | None = None,
    ) -> None:
        """Initialize the DashScope multimodal embedding model class.

        Args:
            api_key (`str`):
                The dashscope API key.
            model_name (`str`):
                The name of the embedding model, e.g. "multimodal-embedding-
                v1", "tongyi-embedding-vision-plus".
            dimensions (`int`, defaults to 1024):
                The dimension of the embedding vector, refer to the
                `official documentation
                <https://bailian.console.aliyun.com/?tab=api#/api/?type=model&url=2712517>`_
                for more details.
            embedding_cache (`EmbeddingCacheBase`):
                The embedding cache class instance, used to cache the
                embedding results to avoid repeated API calls.
        """
        path_doc = (
            "https://bailian.console.aliyun.com/?tab=api#/api/?type=model&"
            "url=2712517"
        )
        self.batch_size_limit = 1

        if model_name.startswith("tongyi-embedding-vision-plus"):
            self.batch_size_limit = 8
            if dimensions is None:
                dimensions = 1152
            elif dimensions != 1152:
                raise ValueError(
                    f"The dimension of model {model_name} must be  1152, "
                    "refer to the official documentation for more details: "
                    f"{path_doc}",
                )
        if model_name.startswith("tongyi-embedding-vision-flash"):
            self.batch_size_limit = 8
            if dimensions is None:
                dimensions = 768
            elif dimensions != 768:
                raise ValueError(
                    f"The dimension of model {model_name} must be  768, "
                    "refer to the official documentation for more details: "
                    f"{path_doc}",
                )
        if model_name.startswith("multimodal-embedding-v"):
            if dimensions is None:
                dimensions = 1024
            elif dimensions != 1024:
                raise ValueError(
                    f"The dimension of model {model_name} must be  1024, "
                    "refer to the official documentation for more details: "
                    f"{path_doc}",
                )
        refined_dimensions: int = 1024
        if dimensions is not None:
            refined_dimensions = dimensions
        super().__init__(model_name, refined_dimensions)

        self.api_key = api_key
        self.embedding_cache = embedding_cache

    async def __call__(
        self,
        inputs: list[TextBlock | ImageBlock | VideoBlock],
        **kwargs: Any,
    ) -> EmbeddingResponse:
        """Call the DashScope multimodal embedding API, which accepts text,
        image, and video data.

        Args:
            inputs (`list[TextBlock | ImageBlock | VideoBlock]`):
                The input data to be embedded. It can be a list of text,
                image, and video blocks.

        Returns:
            `EmbeddingResponse`:
                The embedding response object, which contains the embeddings
                and usage information.
        """
        # check data type
        formatted_data = []
        for _ in inputs:
            if (
                not isinstance(_, dict)
                or "type" not in _
                or _["type"]
                not in [
                    "text",
                    "image",
                    "video",
                ]
            ):
                raise ValueError(
                    f"Invalid data : {_}. It should be a list of "
                    "TextBlock, ImageBlock, or VideoBlock.",
                )
            if (
                _["type"] == "video"
                and _.get("source", {}).get("type") != "url"
            ):
                raise ValueError(
                    f"The multimodal embedding API only supports URL input "
                    f"for video data, but got {_}.",
                )

            if _["type"] == "text":
                assert "text" in _, (
                    f"Invalid text block: {_}. It should contain a "
                    f"'text' field.",
                )
                formatted_data.append({"text": _["text"]})

            elif _["type"] == "video":
                formatted_data.append({"video": _["source"]["url"]})

            elif (
                _["type"] == "image"
                and "source" in _
                and _["source"].get("type") in ["base64", "url"]
            ):
                typ = _["source"]["type"]
                if typ == "base64":
                    formatted_data.append(
                        {
                            "image": f'data:{_["source"]["media_type"]};'
                            f'base64,{_["source"]["data"]}',
                        },
                    )
                elif typ == "url":
                    formatted_data.append(
                        {"image": _["source"]["url"]},
                    )
            else:
                raise ValueError(
                    f"Invalid block {_}. It should be a valid TextBlock, "
                    f"ImageBlock, or VideoBlock.",
                )

        # Handle the batch size limit of the DashScope multimodal embedding API
        collected_embeddings = []
        collected_time = 0.0
        collected_tokens = 0
        collected_source: Literal["cache", "api"] = "cache"
        for _ in range(0, len(formatted_data), self.batch_size_limit):
            batch_data = formatted_data[_ : _ + self.batch_size_limit]
            batch_kwargs = {
                "input": batch_data,
                "model": self.model_name,
                **kwargs,
            }
            res = await self._call_api(batch_kwargs)

            collected_embeddings.extend(res.embeddings)
            collected_time += res.usage.time
            if res.usage.tokens:
                collected_tokens += res.usage.tokens
            if res.source == "api":
                collected_source = "api"

        return EmbeddingResponse(
            embeddings=collected_embeddings,
            usage=EmbeddingUsage(
                tokens=collected_tokens,
                time=collected_time,
            ),
            source=collected_source,
        )

    async def _call_api(self, kwargs: dict[str, Any]) -> EmbeddingResponse:
        """
        Call the DashScope multimodal embedding API by the given arguments.
        """
        # Search in cache first
        if self.embedding_cache:
            cached_embeddings = await self.embedding_cache.retrieve(
                identifier=kwargs,
            )
            if cached_embeddings:
                return EmbeddingResponse(
                    embeddings=cached_embeddings,
                    usage=EmbeddingUsage(
                        tokens=0,
                        time=0,
                    ),
                    source="cache",
                )

        import dashscope

        kwargs["api_key"] = self.api_key

        start_time = datetime.now()
        res = dashscope.MultiModalEmbedding.call(**kwargs)
        time = (datetime.now() - start_time).total_seconds()

        if res.status_code != 200:
            raise RuntimeError(
                f"Failed to get embedding from DashScope API: {res}",
            )

        return EmbeddingResponse(
            embeddings=[_["embedding"] for _ in res.output["embeddings"]],
            usage=EmbeddingUsage(
                tokens=res.usage.get(
                    "image_tokens",
                    0,
                )
                + res.usage.get(
                    "input_tokens",
                    0,
                ),
                time=time,
            ),
            source="api",
        )



# -*- coding: utf-8 -*-
"""The tool response class."""

from dataclasses import dataclass, field
from typing import Optional, List

from .._utils._common import _get_timestamp
from ..message import AudioBlock, ImageBlock, TextBlock


@dataclass
class ToolResponse:
    """The result chunk of a tool call."""

    content: List[TextBlock | ImageBlock | AudioBlock]
    """The execution output of the tool function."""

    metadata: Optional[dict] = None
    """The metadata to be accessed within the agent, so that we don't need to
    parse the tool result block."""

    stream: bool = False
    """Whether the tool output is streamed."""

    is_last: bool = True
    """Whether this is the last response in a stream tool execution."""

    is_interrupted: bool = False
    """Whether the tool execution is interrupted."""

    id: str = field(default_factory=lambda: _get_timestamp(True))
    """The identity of the tool response."""



# -*- coding: utf-8 -*-
"""The session base class in agentscope."""
from abc import abstractmethod

from ..module import StateModule


class SessionBase:
    """The base class for session in agentscope."""

    @abstractmethod
    async def save_session_state(
        self,
        session_id: str,
        **state_modules_mapping: StateModule,
    ) -> None:
        """Save the session state

        Args:
            session_id (`str`):
                The session id.
            **state_modules_mapping (`dict[str, StateModule]`):
                A dictionary mapping of state module names to their instances.
        """

    @abstractmethod
    async def load_session_state(
        self,
        session_id: str,
        allow_not_exist: bool = True,
        **state_modules_mapping: StateModule,
    ) -> None:
        """Load the session state"""



# -*- coding: utf-8 -*-
"""The embedding store base class."""
from abc import abstractmethod
from typing import Any

from .. import Document
from ...types import Embedding


class VDBStoreBase:
    """The vector database store base class, serving as a middle layer between
    the knowledge base and the actual vector database implementation."""

    @abstractmethod
    async def add(self, documents: list[Document], **kwargs: Any) -> None:
        """Record the documents into the vector database."""

    @abstractmethod
    async def delete(self, *args: Any, **kwargs: Any) -> None:
        """Delete texts from the embedding store."""

    @abstractmethod
    async def search(
        self,
        query_embedding: Embedding,
        limit: int,
        score_threshold: float | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        """Retrieve relevant texts for the given queries.

        Args:
            query_embedding (`Embedding`):
                The embedding of the query text.
            limit (`int`):
                The number of relevant documents to retrieve.
            score_threshold (`float | None`, optional):
                The threshold of the score to filter the results.
            **kwargs (`Any`):
                Other keyword arguments for the vector database search API.
        """

    def get_client(self) -> Any:
        """Get the underlying vector database client, so that developers can
        access the full functionality of the vector database."""
        raise NotImplementedError(
            "``get_client`` is not implemented for "
            f"{self.__class__.__name__}.",
        )



# -*- coding: utf-8 -*-
"""The token module in agentscope"""

from ._token_base import TokenCounterBase
from ._gemini_token_counter import GeminiTokenCounter
from ._openai_token_counter import OpenAITokenCounter
from ._anthropic_token_counter import AnthropicTokenCounter
from ._huggingface_token_counter import HuggingFaceTokenCounter

__all__ = [
    "TokenCounterBase",
    "GeminiTokenCounter",
    "OpenAITokenCounter",
    "AnthropicTokenCounter",
    "HuggingFaceTokenCounter",
]



# -*- coding: utf-8 -*-
"""The agent base class in agentscope."""
import asyncio
import io
import json
from asyncio import Task, Queue
from collections import OrderedDict
from copy import deepcopy
from typing import Callable, Any
import base64
import shortuuid
import numpy as np
from typing_extensions import deprecated

from ._agent_meta import _AgentMeta
from .._logging import logger
from ..module import StateModule
from ..message import (
    Msg,
    AudioBlock,
    ToolUseBlock,
    ToolResultBlock,
    ImageBlock,
    VideoBlock,
)
from ..types import AgentHookTypes


class AgentBase(StateModule, metaclass=_AgentMeta):
    """Base class for asynchronous agents."""

    id: str
    """The agent's unique identifier, generated using shortuuid."""

    supported_hook_types: list[str] = [
        "pre_reply",
        "post_reply",
        "pre_print",
        "post_print",
        "pre_observe",
        "post_observe",
    ]
    """Supported hook types for the agent base class."""

    _class_pre_reply_hooks: dict[
        str,
        Callable[
            [
                "AgentBase",  # self
                dict[str, Any],  # kwargs
            ],
            dict[str, Any] | None,  # The modified kwargs or None
        ],
    ] = OrderedDict()
    """The class-level hook functions that will be called before the reply
    function, taking `self` object, the input arguments as input, and
    generating the modified arguments (if needed). Then input arguments of the
    reply function will be re-organized into a keyword arguments dictionary.
    If the one hook returns a new dictionary, the modified arguments will be
    passed to the next hook or the original reply function."""

    _class_post_reply_hooks: dict[
        str,
        Callable[
            [
                "AgentBase",  # self
                dict[str, Any],  # kwargs
                Msg,  # output, the output message
            ],
            Msg | None,
        ],
    ] = OrderedDict()
    """The class-level hook functions that will be called after the reply
    function, which takes the `self` object and deep copied
    positional and keyword arguments (args and kwargs), and the output message
    as input. If the hook returns a message, the new message will be passed
    to the next hook or the original reply function. Otherwise, the original
    output will be passed instead."""

    _class_pre_print_hooks: dict[
        str,
        Callable[
            [
                "AgentBase",  # self
                dict[str, Any],  # kwargs
            ],
            dict[str, Any] | None,  # The modified kwargs or None
        ],
    ] = OrderedDict()
    """The class-level hook functions that will be called before printing,
    which takes the `self` object, a deep copied arguments dictionary as input,
    and output the modified arguments (if needed). """

    _class_post_print_hooks: dict[
        str,
        Callable[
            [
                "AgentBase",  # self
                dict[str, Any],  # kwargs
                Any,  # output, `None` if no output
            ],
            Any,
        ],
    ] = OrderedDict()
    """The class-level hook functions that will be called after the speak
    function, which takes the `self` object as input."""

    _class_pre_observe_hooks: dict[
        str,
        Callable[
            [
                "AgentBase",  # self
                dict[str, Any],  # kwargs
            ],
            dict[str, Any] | None,  # The modified kwargs or None
        ],
    ] = OrderedDict()
    """The class-level hook functions that will be called before the observe
    function, which takes the `self` object and a deep copied input
    arguments dictionary as input. To change the input arguments, the hook
    function needs to output the modified arguments dictionary, which will be
    used as the input of the next hook function or the original observe
    function."""

    _class_post_observe_hooks: dict[
        str,
        Callable[
            [
                "AgentBase",  # self
                dict[str, Any],  # kwargs
                None,  # The output, `None` if no output
            ],
            None,
        ],
    ] = OrderedDict()
    """The class-level hook functions that will be called after the observe
    function, which takes the `self` object as input."""

    def __init__(self) -> None:
        """Initialize the agent."""
        super().__init__()

        self.id = shortuuid.uuid()

        # The replying task and identify of the current replying
        self._reply_task: Task | None = None
        self._reply_id: str | None = None

        # Initialize the instance-level hooks
        self._instance_pre_print_hooks = OrderedDict()
        self._instance_post_print_hooks = OrderedDict()

        self._instance_pre_reply_hooks = OrderedDict()
        self._instance_post_reply_hooks = OrderedDict()

        self._instance_pre_observe_hooks = OrderedDict()
        self._instance_post_observe_hooks = OrderedDict()

        # The prefix used in streaming printing, which will save the
        # accumulated text and audio streaming data for each message id.
        # e.g. {"text": "xxx", "audio": (stream_obj, "{base64_data}")}
        self._stream_prefix = {}

        # The subscribers that will receive the reply message by their
        # `observe` method. The key is the MsgHub id, and the value is the
        # list of agents.
        self._subscribers: dict[str, list[AgentBase]] = {}

        # We add this variable in case developers want to disable the console
        # output of the agent, e.g., in a production environment.
        self._disable_console_output: bool = False

        # The streaming message queue used to export the messages as a
        # generator
        self._disable_msg_queue: bool = True
        self.msg_queue = None

    async def observe(self, msg: Msg | list[Msg] | None) -> None:
        """Receive the given message(s) without generating a reply.

        Args:
            msg (`Msg | list[Msg] | None`):
                The message(s) to be observed.
        """
        raise NotImplementedError(
            f"The observe function is not implemented in"
            f" {self.__class__.__name__} class.",
        )

    async def reply(self, *args: Any, **kwargs: Any) -> Msg:
        """The main logic of the agent, which generates a reply based on the
        current state and input arguments."""
        raise NotImplementedError(
            "The reply function is not implemented in "
            f"{self.__class__.__name__} class.",
        )

    async def print(self, msg: Msg, last: bool = True) -> None:
        """The function to display the message.

        Args:
            msg (`Msg`):
                The message object to be printed.
            last (`bool`, defaults to `True`):
                Whether this is the last one in streaming messages. For
                non-streaming message, this should always be `True`.
        """
        if not self._disable_msg_queue:
            await self.msg_queue.put((deepcopy(msg), last))

        if self._disable_console_output:
            return

        # The accumulated textual content to print, including the text blocks
        # and the thinking blocks
        thinking_and_text_to_print = []

        for block in msg.get_content_blocks():
            if block["type"] == "audio":
                self._process_audio_block(msg.id, block)

            elif block["type"] == "text":
                self._print_text_block(
                    msg.id,
                    name_prefix=msg.name,
                    text_content=block["text"],
                    thinking_and_text_to_print=thinking_and_text_to_print,
                )

            elif block["type"] == "thinking":
                self._print_text_block(
                    msg.id,
                    name_prefix=f"{msg.name}(thinking)",
                    text_content=block["thinking"],
                    thinking_and_text_to_print=thinking_and_text_to_print,
                )

            elif last:
                self._print_last_block(block, msg)

        # Clean up resources if this is the last message in streaming
        if last and msg.id in self._stream_prefix:
            if "audio" in self._stream_prefix[msg.id]:
                player, _ = self._stream_prefix[msg.id]["audio"]
                # Close the miniaudio player
                player.close()
            stream_prefix = self._stream_prefix.pop(msg.id)
            if "text" in stream_prefix and not stream_prefix["text"].endswith(
                "\n",
            ):
                print()

    def _process_audio_block(
        self,
        msg_id: str,
        audio_block: AudioBlock,
    ) -> None:
        """Process audio block content.

        Args:
            msg_id (`str`):
                The unique identifier of the message
            audio_block (`AudioBlock`):
                The audio content block
        """
        if "source" not in audio_block:
            raise ValueError(
                "The audio block must contain the 'source' field.",
            )

        if audio_block["source"]["type"] == "url":
            import urllib.request
            import wave
            import sounddevice as sd

            url = audio_block["source"]["url"]
            try:
                with urllib.request.urlopen(url) as response:
                    audio_data = response.read()

                with wave.open(io.BytesIO(audio_data), "rb") as wf:
                    samplerate = wf.getframerate()
                    n_frames = wf.getnframes()
                    audio_frames = wf.readframes(n_frames)

                    # Convert byte data to numpy array
                    audio_np = np.frombuffer(audio_frames, dtype=np.int16)

                    # Play audio
                    sd.play(audio_np, samplerate)
                    sd.wait()

            except Exception as e:
                logger.error(
                    "Failed to play audio from url %s: %s",
                    url,
                    str(e),
                )

        elif audio_block["source"]["type"] == "base64":
            data = audio_block["source"]["data"]

            if msg_id not in self._stream_prefix:
                self._stream_prefix[msg_id] = {}

            audio_prefix = self._stream_prefix[msg_id].get("audio", None)

            import sounddevice as sd

            # The player and the prefix data is cached for streaming audio
            if audio_prefix:
                player, audio_prefix_data = audio_prefix
            else:
                player = sd.OutputStream(
                    samplerate=24000,
                    channels=1,
                    dtype=np.float32,
                    blocksize=1024,
                    latency="low",
                )
                player.start()
                audio_prefix_data = ""

            # play the audio data
            new_audio_data = data[len(audio_prefix_data) :]
            if new_audio_data:
                audio_bytes = base64.b64decode(new_audio_data)
                audio_np = np.frombuffer(audio_bytes, dtype=np.int16)
                audio_float = audio_np.astype(np.float32) / 32768.0

                # Write to the audio output stream
                player.write(audio_float)

            # save the player and the prefix data
            self._stream_prefix[msg_id]["audio"] = (
                player,
                data,
            )

        else:
            raise ValueError(
                "Unsupported audio source type: "
                f"{audio_block['source']['type']}",
            )

    def _print_text_block(
        self,
        msg_id: str,
        name_prefix: str,
        text_content: str,
        thinking_and_text_to_print: list[str],
    ) -> None:
        """Print the text block and thinking block content.

        Args:
            msg_id (`str`):
                The unique identifier of the message
            name_prefix (`str`):
                The prefix for the message, e.g. "{name}: " for text block and
                "{name}(thinking): " for thinking block.
            text_content (`str`):
                The textual content to be printed.
            thinking_and_text_to_print (`list[str]`):
                A list of textual content to be printed together. Here we
                gather the text and thinking blocks to print them together.
        """
        thinking_and_text_to_print.append(
            f"{name_prefix}: {text_content}",
        )
        # The accumulated text and thinking blocks to print
        to_print = "\n".join(thinking_and_text_to_print)

        # The text prefix that has been printed
        if msg_id not in self._stream_prefix:
            self._stream_prefix[msg_id] = {}

        text_prefix = self._stream_prefix[msg_id].get("text", "")

        # Only print when there is new text content
        if len(to_print) > len(text_prefix):
            print(to_print[len(text_prefix) :], end="")

            # Save the printed text prefix
            self._stream_prefix[msg_id]["text"] = to_print

    def _print_last_block(
        self,
        block: ToolUseBlock | ToolResultBlock | ImageBlock | VideoBlock,
        msg: Msg,
    ) -> None:
        """Process and print the last content block, and the block type
        is not audio, text, or thinking.

        Args:
            block (`ToolUseBlock | ToolResultBlock | ImageBlock | VideoBlock`):
                The content block to be printed
            msg (`Msg`):
                The message object
        """
        text_prefix = self._stream_prefix.get(msg.id, {}).get("text", "")

        if text_prefix:
            # Add a newline to separate from previous text content
            print_newline = "" if text_prefix.endswith("\n") else "\n"
            print(
                f"{print_newline}"
                f"{json.dumps(block, indent=4, ensure_ascii=False)}",
            )
        else:
            print(
                f"{msg.name}:"
                f" {json.dumps(block, indent=4, ensure_ascii=False)}",
            )

    async def __call__(self, *args: Any, **kwargs: Any) -> Msg:
        """Call the reply function with the given arguments."""
        self._reply_id = shortuuid.uuid()

        reply_msg: Msg | None = None
        try:
            self._reply_task = asyncio.current_task()
            reply_msg = await self.reply(*args, **kwargs)

        # The interruption is triggered by calling the interrupt method
        except asyncio.CancelledError:
            reply_msg = await self.handle_interrupt(*args, **kwargs)

        finally:
            # Broadcast the reply message to all subscribers
            if reply_msg:
                await self._broadcast_to_subscribers(reply_msg)
            self._reply_task = None

        return reply_msg

    async def _broadcast_to_subscribers(
        self,
        msg: Msg | list[Msg] | None,
    ) -> None:
        """Broadcast the message to all subscribers."""
        for subscribers in self._subscribers.values():
            for subscriber in subscribers:
                await subscriber.observe(msg)

    async def handle_interrupt(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Msg:
        """The post-processing logic when the reply is interrupted by the
        user or something else."""
        raise NotImplementedError(
            f"The handle_interrupt function is not implemented in "
            f"{self.__class__.__name__}",
        )

    async def interrupt(self, msg: Msg | list[Msg] | None = None) -> None:
        """Interrupt the current reply process."""
        if self._reply_task and not self._reply_task.done():
            self._reply_task.cancel(msg)

    def register_instance_hook(
        self,
        hook_type: AgentHookTypes,
        hook_name: str,
        hook: Callable,
    ) -> None:
        """Register a hook to the agent instance, which only takes effect
        for the current instance.

        Args:
            hook_type (`str`):
                The type of the hook, indicating where the hook is to be
                triggered.
            hook_name (`str`):
                The name of the hook. If the name is already registered, the
                hook will be overwritten.
            hook (`Callable`):
                The hook function.
        """
        if not isinstance(self, AgentBase):
            raise TypeError(
                "The register_instance_hook method should be called on an "
                f"instance of AsyncAgentBase, but got {self} of "
                f"type {type(self)}.",
            )
        hooks = getattr(self, f"_instance_{hook_type}_hooks")
        hooks[hook_name] = hook

    def remove_instance_hook(
        self,
        hook_type: AgentHookTypes,
        hook_name: str,
    ) -> None:
        """Remove an instance-level hook from the agent instance.

        Args:
            hook_type (`AgentHookTypes`):
                The type of the hook, indicating where the hook is to be
                triggered.
            hook_name (`str`):
                The name of the hook to remove.
        """
        if not isinstance(self, AgentBase):
            raise TypeError(
                "The remove_instance_hook method should be called on an "
                f"instance of AsyncAgentBase, but got {self} of "
                f"type {type(self)}.",
            )
        hooks = getattr(self, f"_instance_{hook_type}_hooks")
        if hook_name in hooks:
            del hooks[hook_name]
        else:
            raise ValueError(
                f"Hook '{hook_name}' not found in '{hook_type}' hooks of "
                f"{self.__class__.__name__} instance.",
            )

    @classmethod
    def register_class_hook(
        cls,
        hook_type: AgentHookTypes,
        hook_name: str,
        hook: Callable,
    ) -> None:
        """The universal function to register a hook to the agent class, which
        will take effect for all instances of the class.

        Args:
            hook_type (`AgentHookTypes`):
                The type of the hook, indicating where the hook is to be
                triggered.
            hook_name (`str`):
                The name of the hook. If the name is already registered, the
                hook will be overwritten.
            hook (`Callable`):
                The hook function.
        """

        assert (
            hook_type in cls.supported_hook_types
        ), f"Invalid hook type: {hook_type}"

        hooks = getattr(cls, f"_class_{hook_type}_hooks")
        hooks[hook_name] = hook

    @classmethod
    def remove_class_hook(
        cls,
        hook_type: AgentHookTypes,
        hook_name: str,
    ) -> None:
        """Remove a class-level hook from the agent class.

        Args:
            hook_type (`AgentHookTypes`):
                The type of the hook, indicating where the hook is to be
                triggered.
            hook_name (`str`):
                The name of the hook to remove.
        """

        assert (
            hook_type in cls.supported_hook_types
        ), f"Invalid hook type: {hook_type}"
        hooks = getattr(cls, f"_class_{hook_type}_hooks")
        if hook_name in hooks:
            del hooks[hook_name]

        else:
            raise ValueError(
                f"Hook '{hook_name}' not found in '{hook_type}' hooks of "
                f"{cls.__name__} class.",
            )

    @classmethod
    def clear_class_hooks(
        cls,
        hook_type: AgentHookTypes | None = None,
    ) -> None:
        """Clear all class-level hooks.

        Args:
            hook_type (`AgentHookTypes`, optional):
                The type of the hook to clear. If not specified, all
                class-level hooks will be cleared.
        """

        if hook_type is None:
            for typ in cls.supported_hook_types:
                hooks = getattr(cls, f"_class_{typ}_hooks")
                hooks.clear()
        else:
            assert (
                hook_type in cls.supported_hook_types
            ), f"Invalid hook type: {hook_type}"
            hooks = getattr(cls, f"_class_{hook_type}_hooks")
            hooks.clear()

    def clear_instance_hooks(
        self,
        hook_type: AgentHookTypes | None = None,
    ) -> None:
        """If `hook_type` is not specified, clear all instance-level hooks.
        Otherwise, clear the specified type of instance-level hooks."""
        if hook_type is None:
            for typ in self.supported_hook_types:
                if not hasattr(self, f"_instance_{typ}_hooks"):
                    raise ValueError(
                        f"Call super().__init__() in the constructor "
                        f"to initialize the instance-level hooks for "
                        f"{self.__class__.__name__}.",
                    )
                hooks = getattr(self, f"_instance_{typ}_hooks")
                hooks.clear()

        else:
            assert (
                hook_type in self.supported_hook_types
            ), f"Invalid hook type: {hook_type}"
            if not hasattr(self, f"_instance_{hook_type}_hooks"):
                raise ValueError(
                    f"Call super().__init__() in the constructor "
                    f"to initialize the instance-level hooks for "
                    f"{self.__class__.__name__}.",
                )
            hooks = getattr(self, f"_instance_{hook_type}_hooks")
            hooks.clear()

    def reset_subscribers(
        self,
        msghub_name: str,
        subscribers: list["AgentBase"],
    ) -> None:
        """Reset the subscribers of the agent.

        Args:
            msghub_name (`str`):
                The name of the MsgHub that manages the subscribers.
            subscribers (`list[AgentBase]`):
                A list of agents that will receive the reply message from
                this agent via their `observe` method.
        """
        self._subscribers[msghub_name] = [_ for _ in subscribers if _ != self]

    def remove_subscribers(self, msghub_name: str) -> None:
        """Remove the msghub subscribers by the given msg hub name.

        Args:
            msghub_name (`str`):
                The name of the MsgHub that manages the subscribers.
        """
        if msghub_name not in self._subscribers:
            logger.warning(
                "MsgHub named '%s' not found",
                msghub_name,
            )
        else:
            self._subscribers.pop(msghub_name)

    @deprecated("Please use set_console_output_enabled() instead.")
    def disable_console_output(self) -> None:
        """This function will disable the console output of the agent, e.g.
        in a production environment to avoid messy logs."""
        self._disable_console_output = True

    def set_console_output_enabled(self, enabled: bool) -> None:
        """Enable or disable the console output of the agent. E.g. in a
        production environment, you may want to disable the console output to
        avoid messy logs.

        Args:
            enabled (`bool`):
                If `True`, enable the console output. If `False`, disable
                the console output.
        """
        self._disable_console_output = not enabled

    def set_msg_queue_enabled(
        self,
        enabled: bool,
        queue: Queue | None = None,
    ) -> None:
        """Enable or disable the message queue for streaming outputs.

        Args:
            enabled (`bool`):
                If `True`, enable the message queue to allow streaming
                outputs. If `False`, disable the message queue.
            queue (`Queue | None`, optional):
                The queue instance that will be used to initialize the
                message queue when `enable` is `True`.
        """
        if enabled:
            if queue is None:
                if self.msg_queue is None:
                    self.msg_queue = asyncio.Queue(maxsize=100)
            else:
                self.msg_queue = queue
        else:
            self.msg_queue = None

        self._disable_msg_queue = not enabled



# -*- coding: utf-8 -*-
"""The agent base class."""
from ._agent_base import AgentBase
from ._react_agent_base import ReActAgentBase
from ._react_agent import ReActAgent
from ._user_input import (
    UserInputBase,
    UserInputData,
    TerminalUserInput,
    StudioUserInput,
)
from ._user_agent import UserAgent


__all__ = [
    "AgentBase",
    "ReActAgentBase",
    "ReActAgent",
    "UserInputData",
    "UserInputBase",
    "TerminalUserInput",
    "StudioUserInput",
    "UserAgent",
]
